# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=line-too-long
# pylint: disable=unused-import

from dataclasses import dataclass
import re
from typing import Iterable, OrderedDict, Pattern, Sequence, Sized, Any, TypeVar, Union
from collections import OrderedDict
from abc import ABC, abstractmethod


import colemen_utils as c

import volent.settings as _settings
import volent.settings.types as _t
from volent.exceptions import ValidationError
from volent.validate.Validator import Validator


_T = TypeVar("_T")

class PhoneNumber(Validator):

    field_name = None

    default_message = "Invalid phone number"


    def __init__(
        self,
        error:str=None,
        ):
        '''
            A validator that passes if the value is a phone number
            ----------

            Arguments
            -------------------------
            `error` {str}
                The custom error to use if this validator fails.
                Accepts `input`,`field_name`

            Return {any}
            ----------------------
            The test value is successful, raises ValidationError if it fails.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 03-27-2023 09:10:41
            `memberOf`: NoneOf
            `version`: 1.0
            `method_name`: NoneOf
            * @xxx [03-27-2023 09:13:12]: documentation for NoneOf
        '''
        self.regex = re.compile(r"^(?:\+?\d{1,3}[\s.-]?)?\(?\d{3}\)?[\s.-]?\d{3}[\s.-]?\d{4}(?:x?\d{1,5}?)?$")
        # self.regex = re.compile(r"^[\+]?[(]?[0-9]{3}[)]?[-\s\.]?[0-9]{3}[-\s\.]?[0-9]{4,6}$")
        self.error = error or self.default_message  # type: str


    def _format_error(self, value: Union[str,bytes]) -> str:
        args = {
            "input":value,
        }
        if self.field_name is not None:
            args["field_name"] = self.field_name

        return self.error.format(**args)


    def _repr_args(self) -> str:
        return f""

    def __call__(self,value:_T,field_name:str=None)->_T:
        self.field_name = field_name
        if re.search(self.regex,value):
            return value
        # if self.regex.match(value) is None:
        raise ValidationError(self._format_error(value))







# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=line-too-long
# pylint: disable=unused-import


from dataclasses import dataclass
from typing import Iterable,OrderedDict


import colemen_utils as c


import volent.settings.types as _t
import volent.settings as _settings
from volent.Column import Column as _column
from volent.Relationship import Relationship as _relationship
from volent.UniqueConstraint import UniqueConstraint as _uniqueConstraint
from volent.mixins import MySQLGeneratorMixin
from collections import OrderedDict
from volent.exceptions import ValidationError


@dataclass
class NestedField:
    '''A schema field used for retrieving related schemas as a list. '''
    
    
    main:_t._main_type = None
    model:_t.model_type = None
    schema:_t.schema_type = None
    child_schema:_t.schema_type = None
    child_table:str = None
    '''The name of the model that this field references.'''


    name:str = None
    '''The name of this nested field.'''

    _description:str = None
    column:_t.column_type = None


    required:bool = False
    nullable:bool = True
    default = None
    validators = None


    def __init__(
        self,
        child_table:str=None,
        schema:_t.schema_type=None,
        default=_settings.types.no_default,
        ):
        # '''
        #     Create a schema Field
        #     ----------

        #     Arguments
        #     -------------------------
        #     [`column`=None] {str}
        #         The name of the column that this field represents.
        #         The dot delimited path to the column

        #         If None, it will attempt to find a matching column in the model.

        #     [`required`=False] {bool}
        #         arg_description

        #     [`empty_string_is_null`=True] {bool}
        #         Treat empty strings as None

        #     [`default`] {any}
        #         The default value to assign to this field

        #     [`validate`=None] {any}
        #         A list of validators to apply to this field

        #     Meta
        #     ----------
        #     `author`: Colemen Atwood
        #     `created`: 03-26-2023 09:34:14
        #     `memberOf`: Field
        #     `version`: 1.0
        #     `method_name`: Field
        #     * @xxx [04-14-2023 08:24:58]: documentation for Field
        # '''
        self.child_table = child_table
        self.child_schema = schema
        self.default = default
        # self.required = required
        # self.nullable = nullable
        # self.empty_string_is_null = empty_string_is_null

        # self.validators = c.arr.force_list(validate,allow_nulls=False)


    @property
    def summary(self):
        '''
            Get this Model's summary

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 03-23-2023 14:39:08
            `@memberOf`: Model
            `@property`: summary
        '''
        value = {
            "name":self.name,
        }
        return value

    @property
    def dump_only(self):
        '''
            Get this Field's dump_only

            If True, this field can only be retrieved from the database and not inserted or updated.

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 04-18-2023 15:42:42
            `@memberOf`: Field
            `@property`: dump_only
        '''
        value = self.column.dump_only
        return value

    @property
    def value(self):
        '''
            Get this Field's value

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 03-25-2023 12:02:50
            `@memberOf`: Field
            `@property`: value
        '''
        value = self.column.value
        return value

    def validate(self):
        # val = self.value
        val = self.column.deserialized_value
        # val = self.column.data_type.

        if isinstance(val,self.column.data_type.python_data_type) is False:
            if self.column.is_primary is True:
                pass
            elif self.column.nullable is True:
                pass
            else:
                raise ValidationError(f"{self.column.name} expects {self.column.data_type.python_data_type} types.",self.name)

        if self._less_than_data_len(val) is False:
            raise ValidationError(f"{self.column.name} is too long.",self.name)



        if self._is_null(val):
            raise ValidationError(f"{self.name} cannot be null.",self.name)


        for valid in self.validators:
            val = valid(val,self.name)

    def _is_null(self,val):
        if isinstance(val,(str)):
            if len(val) == 0:
                val = None

        if self.nullable is False and self.value is None:
            return True
        return False

    def _less_than_data_len(self,val):
        if isinstance(self.column.data_type.data_length,(int)):
            val = str(val)
            length = len(val)
            if length > self.column.data_type.data_length:
                return False
        return True




    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} : {self.name}>"



    def open_api_definition(self):
        '''
            Generate the open API definition for this schema.

            ----------

            Return {dict}
            ----------------------
            A dictionary definition of this schema.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 05-07-2023 15:30:49
            `memberOf`: Schema
            `version`: 1.0
            `method_name`: open_api_definition
            * @xxx [05-07-2023 15:32:48]: documentation for open_api_definition
        '''
        value = {}
        for f in self.child_schema().fields:
            value[f.name] = f.open_api_definition()
        if hasattr(self.child_schema(),"nested_fields"):
            nest = self.child_schema().nested_fields
            if nest is not None:
                for f in nest:
                    value[f.name] = f.open_api_definition()

        return value

    def open_api_data(self,loc:str="body")->Iterable[dict]:
        '''
            Get this Schema's open_api_data

            returns a list of dictionaries each representing a field, the dictionaries can
            be used with the open api library for documentation.

            ----------

            Arguments
            -------------------------
            [`loc`='body'] {str}
                Where the field's value can be found in a request ['body','path']


            Return {list}
            ----------------------
            A list of dictionaries.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 04-24-2023 09:17:49
            `memberOf`: Schema
            `version`: 1.0
            `method_name`: open_api_data
            `throws`: ValueError - if the loc is not "body"/"path"
            * @xxx [04-24-2023 09:19:05]: documentation for open_api_data
        '''

        value = []
        if loc not in _settings.control.open_api_param_locations:
            raise ValueError(f"{loc} is not a valid param location, expected: {','.join(_settings.control.open_api_param_locations)}")
        for f in self.fields:
            value.append(f.open_api_data(loc))
        return value


# def _gather_subs(model:_t.model_type):
#     columns = []
#     relationships:Iterable[_t.relationship_type] = []


#     # @Mstep [] create an base instance to compare to.
#     # print(f"model.__dataclass_fields__:{model.__fields__}")
#     # print(list(model.__dict__.keys()))
#     # print("\n\n\n")
#     # print(dir(Model()))
#     df_props = dir(Model())
#     # @Mstep [] gather the props of this instance.
#     props = dir(model)
#     # @Mstep [] find the props that exist on this instance and not on the base.
#     dif = c.arr.find_list_diff(props,df_props)

#     for prop in dif:
#         # print(f"prop: {prop} - {type(prop)}")
#         if isinstance(getattr(model,prop),_column):
#             prop:_t.model_type = getattr(model,prop)
#             prop.database = model.database
#             prop.main = model.main
#             prop.model = model
#             columns.append(prop)
#             continue

#         if isinstance(getattr(model,prop),_relationship):
#             prop:_t.model_type = getattr(model,prop)
#             prop.database = model.database
#             prop.main = model.main
#             prop.child_model = model
#             if isinstance(prop.child,_column):
#                 prop.child_column = prop.child
#             if isinstance(prop,(str)):
#                 prop.child_column = model.get_column(prop)
#             relationships.append(prop)

#     order_cols = []
#     for k in list(model.__fields__):
#         for col in columns:
#             if col.name == k:
#                 order_cols.append(col)
#         # if k in columns:
#         #     order_cols.append(columns[])



#     model._columns = order_cols
#     model._relationships = relationships








# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=line-too-long
# pylint: disable=unused-import

from dataclasses import dataclass
import re
from typing import Iterable, OrderedDict, Pattern, Sequence, Sized, Any, TypeVar, Union
from collections import OrderedDict
from abc import ABC, abstractmethod


import colemen_utils as c

import volent.settings as _settings
import volent.settings.types as _t
from volent.exceptions import ValidationError
from volent.validate.Validator import Validator


_T = TypeVar("_T")

class IpAddress(Validator):

    field_name = None

    default_message = "Invalid IP Address"


    def __init__(
        self,
        ipv4:bool=True,
        ipv6:bool=True,
        error:str=None,
        ):
        '''
            A validator that passes if the value is an ip address
            ----------

            Arguments
            -------------------------
            [`ipv4`=True] {bool}
                If True ipv4 ip addresses are allowed
            [`ipv6`=True] {bool}
                If True ipv6 ip addresses are allowed

            `error` {str}
                The custom error to use if this validator fails.
                Accepts `input`,`field_name`

            Return {any}
            ----------------------
            The test value is successful, raises ValidationError if it fails.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 03-27-2023 09:10:41
            `memberOf`: NoneOf
            `version`: 1.0
            `method_name`: NoneOf
            * @xxx [03-27-2023 09:13:12]: documentation for NoneOf
        '''
        self.ipv4 = ipv4
        self.ipv6 = ipv6
        # self.ipv4_regex = re.compile(r"^(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)$")
        self.ipv4_regex = re.compile(r"^((25[0-5]|2[0-4][0-9]|1[0-9][0-9]|[1-9]?[0-9])\.){3}(25[0-5]|2[0-4][0-9]|1[0-9][0-9]|[1-9]?[0-9])$")
        # self.ipv4_regex = re.compile(r"(?:[0-9]{1,3}\.){3}[0-9]{1,3}")
        ipv6_regex = "((([0-9a-fA-F]){1,4})\\:){7}"\
            "([0-9a-fA-F]){1,4}"
        self.ipv6_regex = re.compile(ipv6_regex)
        # self.ipv6_regex = re.compile(r"(([0-9a-fA-F]{1,4}:){7,7}[0-9a-fA-F]{1,4}|([0-9a-fA-F]{1,4}:){1,7}:|([0-9a-fA-F]{1,4}:){1,6}:[0-9a-fA-F]{1,4}|([0-9a-fA-F]{1,4}:){1,5}(:[0-9a-fA-F]{1,4}){1,2}|([0-9a-fA-F]{1,4}:){1,4}(:[0-9a-fA-F]{1,4}){1,3}|([0-9a-fA-F]{1,4}:){1,3}(:[0-9a-fA-F]{1,4}){1,4}|([0-9a-fA-F]{1,4}:){1,2}(:[0-9a-fA-F]{1,4}){1,5}|[0-9a-fA-F]{1,4}:((:[0-9a-fA-F]{1,4}){1,6})|:((:[0-9a-fA-F]{1,4}){1,7}|:)|fe80:(:[0-9a-fA-F]{0,4}){0,4}%[0-9a-zA-Z]{1,}|::(ffff(:0{1,4}){0,1}:){0,1}((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}(25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])|([0-9a-fA-F]{1,4}:){1,4}:((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}(25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9]))")

        self.error = error or self.default_message  # type: str


    def _format_error(self, value: Union[str,bytes]) -> str:
        args = {
            "input":value,
        }
        if self.field_name is not None:
            args["field_name"] = self.field_name

        return self.error.format(**args)


    def _repr_args(self) -> str:
        return f"ipv4={self.ipv4!r},ipv6={self.ipv6!r}"

    def __call__(self,value:_T,field_name:str=None)->_T:
        self.field_name = field_name
        if isinstance(value,(str,bytes)) is False:
            return value
        if self.ipv4 is True:
            if re.search(self.ipv4_regex,value):
            # if self.ipv4_regex.match(value) is True:
                return value

        if self.ipv6 is True:
            if re.search(self.ipv6_regex,value):
            # if self.ipv6_regex.match(value) is True:
                return value

        raise ValidationError(self._format_error(value))





# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=line-too-long
# pylint: disable=unused-import

from dataclasses import dataclass
from typing import Iterable, OrderedDict, Sequence, Sized, Any, TypeVar, Union
from collections import OrderedDict
from abc import ABC, abstractmethod


import colemen_utils as c

import volent.settings as _settings
import volent.settings.types as _t
from volent.exceptions import ValidationError
from volent.validate.Validator import Validator


_T = TypeVar("_T")

class Range(Validator):

    field_name = None
    

    message_min = "Must be {min_op} {{min}}."
    message_max = "Must be {max_op} {{max}}."
    message_all = "Must be {min_op} {{min}} and {max_op} {{max}}."

    message_gte = "greater than or equal to"
    message_gt = "greater than"
    message_lte = "less than or equal to"
    message_lt = "less than"
    
    
    # default_message = "Must be within range {min}-{max}"
    # below_min_message = "Must be greater than {min}"
    # above_max_message = "Must be less than {max}"


    def __init__(
        self,
        min:Union[int,float]=None,
        max:Union[int,float]=None,
        min_inclusive:bool=True,
        max_inclusive:bool=True,
        error:str=None,
        ):
        '''
            A validator that passes if the value is within the range.
            ----------

            Arguments
            -------------------------
            `min` {int,float}
                The minimum value

            `max` {int,float}
                The maximum value

            `min_inclusive` {bool}
                If False, the min value is not considered a valid value

            `max_inclusive` {bool}
                If False, the max value is not considered a valid value

            `error` {str}
                The custom error to use if this validator fails.
                Accepts `input`, `min`, `max`, `field_name`

            Return {any}
            ----------------------
            The test value is successful, raises ValidationError if it fails.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 03-27-2023 09:10:41
            `memberOf`: NoneOf
            `version`: 1.0
            `method_name`: NoneOf
            * @xxx [03-27-2023 09:13:12]: documentation for NoneOf
        '''
        self.min = min
        self.max = max
        self.min_inclusive = min_inclusive
        self.max_inclusive = max_inclusive
        self.error = error
        # self.use_custom_error = False
        # if error is not None:
            # self.use_custom_error = True
        self.message_min = self.message_min.format(
            min_op=self.message_gte if self.min_inclusive else self.message_gt
        )
        self.message_max = self.message_max.format(
            max_op=self.message_lte if self.max_inclusive else self.message_lt
        )
        self.message_all = self.message_all.format(
            min_op=self.message_gte if self.min_inclusive else self.message_gt,
            max_op=self.message_lte if self.max_inclusive else self.message_lt,
        )
        # self.error = error or self.default_message  # type: str

        if min is None and max is None:
            raise ValueError("min or max must be provided.")


    def _format_error(self, value: _T,message:str) -> str:
        args = {
            "input":value,
            "min":self.min,
            "max":self.max,
        }
        if self.field_name is not None:
            args["field_name"] = self.field_name

        return (self.error or message).format(**args)


    def _repr_args(self) -> str:
        return f"min={self.min!r},max={self.max!r},min_inclusive={self.min_inclusive!r},max_inclusive={self.max_inclusive!r}"


    def __call__(self,value:_T,field_name:str=None)->_T:
        self.field_name = field_name
        if value is None:
            return value
        try:
            if self.min is not None and (
                value < self.min if self.min_inclusive else value <= self.min
            ):
                message = self.message_min if self.max is None else self.message_all
                raise ValidationError(self._format_error(value, message))

            if self.max is not None and (
                value > self.max if self.max_inclusive else value >= self.max
            ):
                message = self.message_max if self.min is None else self.message_all
                raise ValidationError(self._format_error(value, message))
        except TypeError as e:
            raise ValidationError(f"{field_name} - Cannot compare {type(value)} {e}")

            
        return value

        return value


# class ContainsOnly(OneOf):

#     field_name = None
#     default_message = "One or more of the choices you made was not in: {choices}."


#     def __init__(
#         self,
#         choices=Iterable,
#         ):

#         self.choices = choices
#         self.choices_str = ', '.join(str(choice) for choice in self.choices)

#     def _format_error(self, value: str, message: str) -> str:
#         if self.field_name is not None:
#             message = f"{self.field_name} was not in {self.choices}"
#         return message.format(
#             input=value, choices=self.choices_str
#         )


#     def _repr_args(self) -> str:
#         return f"min={self.min!r}, max={self.max!r}, equal={self.equal!r}"


#     def __call__(self,value:Sequence[_T],field_name:str=None)->Sequence[_T]:
#         self.field_name = field_name
#         test_value = c.arr.force_list(value)

#         for val in test_value:
#             if val not in self.choices:
#                 raise ValidationError(self._format_error(value,self.default_message))

#         return value



# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=line-too-long
# pylint: disable=unused-import



class EntityName(type):

    @property
    def name(self):
        '''
            Get this EntityName's name

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 04-11-2023 14:32:58
            `@memberOf`: EntityName
            `@property`: name
        '''
        value = self._name
        if value is None:
            if hasattr(self,"__table_name__"):
                value = getattr(self,"__table_name__")
            self._name = value
        return value


# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=line-too-long
# pylint: disable=unused-import
'''
    The query delete module.

    ----------

    Meta
    ----------
    `author`: Colemen Atwood
    `created`: 04-28-2023 07:24:17
    `name`: delete
    * @xxx [04-28-2023 07:24:33]: documentation for delete
'''



from dataclasses import dataclass
from typing import Iterable, Union


import colemen_utils as c


import volent.settings.types as _t
import volent.settings as _settings
# from volent.Field import Field as _field
# from volent.Relationship import Relationship as _relationship
# from volent.UniqueConstraint import UniqueConstraint as _uniqueConstraint


@dataclass
class QueryParam:
    Query:_t.query_type = None
    name:str = None
    value:any = None

    def __init__(self,Query:_t.query_type,name:str,value) -> None:
        self.Query = Query
        self.name = name
        self.value = value

    def __repr__(self) -> str:
        return f"<QueryParam : {self.name} :: {self.value}>"

    def serialize(self)->dict:
        '''Get this param as a dictionary.'''
        return {self.name:self.value}

    @property
    def parameterized_name(self)->str:
        '''Generate the parameterized name of this param.

        Example:
        - @title : when sqlite is being used.'''
        if _settings.globe.flavor in ["sqlite"]:
            return f"@{self.name}"
        else:
            return f":{self.name}"
    # parameterized_name = paramaterized_name

    @property
    def parameterized_update_string(self)->str:
        '''Generate the placeholder string used for this parameter when updating a column.

        Example:
        - track_number = @track_number
        '''
        if _settings.globe.flavor in ["sqlite"]:
            return f"{self.name}=@{self.name}"
        else:
            return f"{self.name}=:{self.name}"




"""API tests"""
# pylint: disable=invalid-name

from dataclasses import dataclass
from datetime import datetime
from datetime import timezone


import colemen_utils as c
from .. import volent

# import volent.Volent as volent
# from volent.Volent import Schema,Field,validators as _v
# from volent.data_types import String,Integer,EncodedPrimary




# @dataclass
# class SuccessfulSchema(Schema):
#     '''The publicly visible data for a task'''


#     def __init__(
#         self,
#         ) -> None:
#         # self.task_id = Field(required=True,nullable=False)
#         # self.target_id = Field(data_type=EncodedPrimary())
#         self.limit = Field(data_type=Integer(),default=50,validate=[_v.OneOf([25,50,100])],description="The number of results to return.")
#         self.offset = Field(data_type=Integer(),default=0,validate=[_v.Range(0)])
#         self.start_timestamp = Field(data_type=Integer(),default=None)
#         self.end_timestamp = Field(data_type=Integer(),default=None)

#     def valid_timestamps(self):
#         # print(f"validating schema timestamps")
#         s = self.start_timestamp.value
#         e = self.end_timestamp.value

#         if s is not None and e is not None:
#             if e > s:
#                 raise ValueError("Start timestamp must be after the end timestamp")







class TestSchemaDict:
    """TEST Schema Validation of dictionaries"""

    def test_successful_schema(self):
        """Confirm that we can create a new user"""
        assert True

        # v = volent.Volent()

        # data = {
        #     # "target_id":c.string.string_encode_int(50),
        #     # "target_id":123456,
        #     "limit":50,
        #     "offset":50,
        #     # "startTimestamp":round(datetime.now(tz=timezone.utc).timestamp()) - (86400 * 8),
        #     "startTimestamp":500,
        #     "endTimestamp":50,
        # }
        # ts = SuccessfulSchema()
        # dumped = ts.dump(data)
        # assert isinstance(dumped,(dict)) is True



# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=line-too-long
# pylint: disable=unused-import

from dataclasses import dataclass
from string import Template
from typing import Iterable, Union
# from typing import Iterable

import colemen_utils as c
import sqlite3
import mysql.connector as _mysqlConnnector


import volent.settings as _settings
import volent.settings.types as _t

qc:str = _settings.control.sql_quote_char
'''The quote character to use.'''
indent:str = " " * _settings.control.mysql.create_table_indent
'''The indentation string used for the table contents'''


@dataclass
class MySQLGeneratorMixin:


    def master_sql(self,file_path:str=None,drops=True):
        '''Generate the Master SQL for all databases and tables.

        This can only be called on the Volent instance.,
        '''
        database_drops = []
        database_creates = []
        table_drops = []
        table_creates = []
        master_string = ""

        from volent.Volent import Volent as _bv
        from volent.Database import Database as _database
        if isinstance(self,_bv):

            # print(f"generating MYSQL SQL content")
            for mdl in self.models:
                table_drops.append(mdl.drop_statement)
                table_creates.append(mdl.create_statement)
            for db in self.databases:
                database_drops.append(db.drop_statement)
                database_creates.append(db.create_statement)
            if drops:
                master_list = table_drops + database_drops + database_creates + table_creates
            else:
                master_list = database_creates + table_creates
            master_string = '\n'.join(master_list)


        if isinstance(self,_database):

            for mdl in self.models:
                table_drops.append(mdl.drop_statement)
                table_creates.append(mdl.create_statement)
            database_drops.append(self.drop_statement)
            database_creates.append(self.create_statement)
            if drops:
                master_list = table_drops + database_drops + database_creates + table_creates
            else:
                master_list = database_creates + table_creates
            master_string = '\n'.join(master_list)

        c.file.write(file_path,master_string)
        return master_string

    @property
    def quoted_name(self):
        '''
            Get this MySQLGeneratorMixin's quoted_name

            If this is a database:
                \`database_name\`
            If this is a model:
                \`database_name\`.\`table_name\`

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 03-25-2023 09:35:15
            `@memberOf`: MySQLGeneratorMixin
            `@property`: quoted_name
        '''
        from volent.Database import Database as _database
        from volent.Model import Model as _model
        from volent.Column import Column as _column
        from volent.UniqueConstraint import UniqueConstraint as _unique_constraint
        from volent.Relationship import Relationship as _relationship

        if isinstance(self,_database):
            return f"{qc}{self.name}{qc}"

        if isinstance(self,_model):
            if self.database_name is not None and _settings.globe.flavor == "mysql":
                # print(self.database.name)
                # print(self.database.name.name)
                return f"{qc}{self.database_name}{qc}.{qc}{self.model_name}{qc}"
            else:
                return f"{qc}{self.model_name}{qc}"

        if isinstance(self,_column):
            return f"{qc}{self.name}{qc}"

        if isinstance(self,_unique_constraint):
            return f"{qc}{self.name}{qc}"

        if isinstance(self,_relationship):
            return f"{qc}{self.name}{qc}"

    def gen_quoted_name(self,table_name:str,database_name:str=None):
        if isinstance(database_name,(str)):
            database_name = f"{qc}{database_name}{qc}."
        else:
            database_name = ""
        value = f"{database_name}{qc}{table_name}{qc}"
        return value

    @property
    def drop_statement(self):
        from volent.Database import Database as _database
        from volent.Model import Model as _model

        if isinstance(self,_database):
            exists = _gen_safe_script_exists()
            return f"DROP SCHEMA {exists}{self.quoted_name};"
        if isinstance(self,_model):
            exists = _gen_safe_script_exists()
            return f"DROP TABLE {exists}{self.quoted_name};"

    def gen_drop_table_statement(self,table_name,database_name:str=None):
        if _settings.globe.flavor in ['sqlite']:
            value = f"DROP TABLE IF EXISTS {_settings.control.sql_quote_char}{table_name}{_settings.control.sql_quote_char};"
            return value
            database_name = None
        qn = self.gen_quoted_name(table_name,database_name)
        value = f"DROP TABLE IF EXISTS {qn};DROP TABLE IF EXISTS {qn}_fts"
        # value = f"DROP TABLE IF EXISTS {qn}"
        return value


    @property
    def create_statement(self):
        from volent.Database import Database as _database
        from volent.Model import Model as _model
        from volent.Column import Column as _column
        from volent.UniqueConstraint import UniqueConstraint as _unique_constraint
        from volent.FullTextIndex import FullTextIndex as _full_text_index
        from volent.Relationship import Relationship as _relationship


        if isinstance(self,_database):
            exists = _gen_safe_script_not_exists()
            return f"CREATE SCHEMA {exists}{self.quoted_name};"

        if isinstance(self,_column):
            return _gen_column_declaration(self)

        if isinstance(self,_unique_constraint):
            if _settings.globe.flavor in ["sqlite"]:
                return _gen_sqlite_unique_constraint_create(self)
            return _gen_unique_constraint_create(self)

        if isinstance(self,_full_text_index):
            if _settings.globe.flavor in ['sqlite']:
                return None
            return _gen_full_text_index_create(self)

        if isinstance(self,_relationship):
            if _settings.globe.flavor in ["sqlite"]:
                return _gen_sqlite_relationship_create(self)
            return _gen_relationship_create(self)

        if isinstance(self,_model):
            return _gen_table_create_statement(self)

    @property
    def show_columns(self):
        '''
            Get this MySQLGeneratorMixin's show_columns

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 04-18-2023 09:36:19
            `@memberOf`: MySQLGeneratorMixin
            `@property`: show_columns
        '''
        
        from volent.Model import Model as _model
        if isinstance(self,_model):
            sql = f"SHOW FULL COLUMNS FROM {self.quoted_name}"
            return sql

    @property
    def describe(self):
        '''
            Get the 

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 04-18-2023 09:33:48
            `@memberOf`: MySQLGeneratorMixin
            `@property`: describe
        '''
        from volent.Model import Model as _model
        if isinstance(self,_model):
            sql = f"DESC {self.quoted_name}"
            return sql







# ---------------------------------------------------------------------------- #
#                               COLUMN GENERATORS                              #
# ---------------------------------------------------------------------------- #


def _gen_column_declaration(column:_t.column_type)->str:
    '''Generate a mysql column declaration

    `device_id`      bigint NOT NULL COMMENT 'description: The id of the device to associate'
    '''
    value = []
    lcn,_ = column.model.longest_column_name
    q = _settings.control.sql_quote_char
    sql_name = f"{q}{column.name}{q}"
    sql_name = c.string.rightPad(sql_name,lcn + 5," ")
    data_type = column.data_type
    if isinstance(column.data_type,(str)) is False:
        data_type = column.data_type.sql

    auto_increment = _gen_auto_increment(column.auto_increment)
    nullable = _gen_sql_nullable(column.nullable)
    comment = _gen_column_comment(column.comment)
    default = _gen_column_default(column)
    primary_key = _gen_column_primary_key(column)
    value.append(sql_name)
    value.append(data_type)
    value.append(primary_key)
    value.append(nullable)
    value.append(auto_increment)
    value.append(default)
    value.append(comment)
    value = c.arr.strip_list_nulls(value)
    string_value = ' '.join(value)
    indent = " " * _settings.control.mysql.create_table_indent
    return f"{indent}{string_value}"

def _gen_column_primary_key(column:_t.column_type)->str:
    if _settings.globe.flavor in ['sqlite']:
        if column.is_primary is True:
            return "PRIMARY KEY"
    return None

def _gen_sql_nullable(nullable:bool)->str:
    '''
        Generates the MySQL null value

        "NULL" or "NOT NULL"
        ----------

        Arguments
        -------------------------
        `nullable` {bool}
            The nullable value of the column

        Return {str}
        ----------------------
        "NULL" if the column is nullable, "NOT NULL" if it is not nullable

        Meta
        ----------
        `author`: Colemen Atwood
        `created`: 03-24-2023 06:33:06
        `version`: 1.0
        `method_name`: _gen_sql_nullable
        * @xxx [03-24-2023 06:34:55]: documentation for _gen_sql_nullable
    '''
    value = "NOT NULL"
    if nullable:
        value = "NULL"

    return value

def _gen_auto_increment(auto_increment:bool)->str:
    
    value = None
    if auto_increment and _settings.globe.flavor == "mysql":
        value = "AUTO_INCREMENT"
    # if auto_increment and _settings.globe.flavor in ["sqlite"]:
    #     value = "AUTOINCREMENT"

    return value

def _gen_column_comment(comment:str=None)->str:
    '''
        Generate a MySQL column comment
        ----------

        Arguments
        -------------------------
        `comment` {str,None}
            The comment value


        Return {str,None}
        ----------------------
        The comment string if the value is not None and has a length greater than zero

        Meta
        ----------
        `author`: Colemen Atwood
        `created`: 03-24-2023 06:39:39
        `version`: 1.0
        `method_name`: _gen_column_comment
        * @TODO []: documentation for _gen_column_comment
    '''
    value = None
    if _settings.globe.flavor in ['sqlite']:
        return None
    if comment is None:
        return comment
    comment = c.string.strip(comment,[" "])
    if len(comment) == 0:
        return value


    return f"COMMENT '{comment}'"

def _gen_column_default(column:_t.column_type)->str:
    '''
        Generate a MySQL column default
        ----------

        Arguments
        -------------------------
        `comment` {str,None}
            The comment value


        Return {str,None}
        ----------------------
        The comment string if the value is not None and has a length greater than zero

        Meta
        ----------
        `author`: Colemen Atwood
        `created`: 03-24-2023 06:39:39
        `version`: 1.0
        `method_name`: _gen_column_comment
        * @TODO []: documentation for _gen_column_comment
    '''
    value = column.default
    # dt = column.data_type.python_data_type
    dt = column.data_type.python_type_name
    # print(f"dt:{dt}")
    if value is None:
        return "DEFAULT NULL"

    if isinstance(value,_settings.types.NoDefault):
        return None


    if dt in ["str","string"]:
        return f"DEFAULT '{value}'"
    if dt in ["bool"]:
        return f"DEFAULT {c.types.bool_to_int(value)}"
    if dt in ["int"]:
        return f"DEFAULT {int(value)}"
    if dt in ["float"]:
        return f"DEFAULT {float(value)}"


    return None




# ---------------------------------------------------------------------------- #
#                               UNIQUE CONSTRAINT                              #
# ---------------------------------------------------------------------------- #


_UNIQUE_CONSTRAINT_TEMPLATE = '''${indent}UNIQUE KEY $name ($columns) $comment'''
def _gen_unique_constraint_create(u:_t.unique_constraint_type)->str:
    s = Template(_UNIQUE_CONSTRAINT_TEMPLATE)
    # col_list = []
    # for col in u.columns:
    #     col_list.append(col.quoted_name)
    # columns = ','.join(col_list)
    columns = __quoted_col_name_list(u)
    # comment = _gen_column_comment(u.comment) or ""
    # if len(comment) > 0:
    #     comment = f" {comment}"
    val = s.substitute(
        indent=indent,
        name=u.quoted_name,
        columns=columns,
        comment=_gen_column_comment(u.comment) or ""
    )
    val = c.string.strip(val,[" "],"right")
    return val

_SQLITE_UNIQUE_CONSTRAINT_TEMPLATE = '''${indent}UNIQUE($columns)'''
def _gen_sqlite_unique_constraint_create(u:_t.unique_constraint_type)->str:
    s = Template(_SQLITE_UNIQUE_CONSTRAINT_TEMPLATE)
    # col_list = []
    # for col in u.columns:
    #     col_list.append(col.quoted_name)
    # columns = ','.join(col_list)
    columns = __quoted_col_name_list(u)
    # comment = _gen_column_comment(u.comment) or ""
    # if len(comment) > 0:
    #     comment = f" {comment}"
    val = s.substitute(
        indent=indent,
        name=u.quoted_name,
        columns=columns,
    )
    val = c.string.strip(val,[" "],"right")
    return val

# ---------------------------------------------------------------------------- #
#                               INDEX GENERATORS                               #
# ---------------------------------------------------------------------------- #

_FULLTEXT_TEMPLATE = '''${indent}FULLTEXT ($columns)'''
def _gen_full_text_index_create(ft:_t.full_text_index_type)->str:
    s = Template(_FULLTEXT_TEMPLATE)
    # col_list = []
    # for col in ft.columns:
    #     col_list.append(col.quoted_name)
    # columns = ','.join(col_list)
    columns = __quoted_col_name_list(ft)

    val = s.substitute(
        indent=indent,
        columns=columns,
    )
    return val

def __quoted_col_name_list(entity:Union[_t.unique_constraint_type,_t.full_text_index_type]):
    col_list = []
    for col in entity.columns:
        col_list.append(col.quoted_name)
    columns = ','.join(col_list)
    return columns

# ---------------------------------------------------------------------------- #
#                            FOREIGN KEY GENERATORS                            #
# ---------------------------------------------------------------------------- #


_RELATIONSHIP_TEMPLATE = '''${indent}KEY ${qc}$fk_name${qc} ($child_column_name),
${indent}CONSTRAINT $relationship_name FOREIGN KEY ${qc}$fk_name${qc} ($child_column_name) REFERENCES $parent_table ($parent_column)$on_update$on_delete'''

def _gen_relationship_create(r:_t.relationship_type)->str:
    on_update = ""
    on_delete = ""
    if r.on_update is not None:
        on_update = f" ON UPDATE {r.on_update}"
    if r.on_delete is not None:
        on_delete = f" ON DELETE {r.on_delete}"
    s = Template(_RELATIONSHIP_TEMPLATE)
    december= s.substitute(
        indent=indent,
        qc=qc,
        # fk_name=f"FK_{c.rand.rand()}",
        fk_name=r.fk_name,
        child_column_name=r.child_column.quoted_name,
        relationship_name=r.quoted_name,
        parent_table=r.parent_model.quoted_name,
        parent_column=r.parent_column.quoted_name,
        on_delete=on_delete,
        on_update=on_update,
    )
    return december



_SQLITE_RELATIONSHIP_TEMPLATE = '''${indent}FOREIGN KEY($child_column_name) REFERENCES $parent_table($parent_column)'''

def _gen_sqlite_relationship_create(r:_t.relationship_type)->str:
    on_update = ""
    on_delete = ""
    if r.on_update is not None:
        on_update = f" ON UPDATE {r.on_update}"
    if r.on_delete is not None:
        on_delete = f" ON DELETE {r.on_delete}"
    s = Template(_SQLITE_RELATIONSHIP_TEMPLATE)
    december= s.substitute(
        indent=indent,
        qc=qc,
        # fk_name=f"FK_{c.rand.rand()}",
        fk_name=r.fk_name,
        child_column_name=r.child_column.quoted_name,
        relationship_name=r.quoted_name,
        parent_table=r.parent_model.quoted_name,
        parent_column=r.parent_column.quoted_name,
        on_delete=on_delete,
        on_update=on_update,
    )
    return december




# ---------------------------------------------------------------------------- #
#                               TABLE GENERATORS                               #
# ---------------------------------------------------------------------------- #


def _gen_table_comment(comment:str=None):
    '''
        Generate a MySQL table comment
        ----------

        Arguments
        -------------------------
        `comment` {str,None}
            The comment value


        Return {str,None}
        ----------------------
        The comment string if the value is not None and has a length greater than zero

        Meta
        ----------
        `author`: Colemen Atwood
        `created`: 03-24-2023 06:39:39
        `version`: 1.0
        `method_name`: _gen_table_comment
        * @TODO []: documentation for _gen_table_comment
    '''
    value = None
    if _settings.globe.flavor in ['sqlite']:
        return ""
    if comment is None:
        return ""
    comment = c.string.strip(comment,[" "])
    if len(comment) == 0:
        return value


    return f" COMMENT='{comment}'"

def _gen_table_create_columns(model:_t.model_type)->Iterable[str]:
    '''Generates a list of column create statements for a table'''
    columns = []
    for col in model.columns:
        columns.append(col.create_statement)
    return columns

def _gen_primary_columns(model:_t.model_type)->Iterable[str]:
    '''Generates a list of primary key statements for a table'''
    if _settings.globe.flavor in ["sqlite"]:
        return []
    primary_cols = model.primary_columns
    if len(primary_cols) == 1:
        col:_t.column_type = primary_cols[0]
        return [f"{indent}PRIMARY KEY({col.quoted_name})"]

def _gen_unique_constraints(model:_t.model_type)->Iterable[str]:
    '''Generates a list of unique constraint statements for a table'''
    uqs = []
    uc = model.unique_constraints
    if isinstance(uc,(list)):
        for u in uc:
            uqs.append(u.create_statement)
    return uqs

def _gen_fulltext_indexes(model:_t.model_type)->Iterable[str]:
    '''Generates a list of fulltext index statements for a table'''
    ftis = []
    fts = model.full_text_indexes
    if isinstance(fts,(list)):
        for ft in fts:
            ftis.append(ft.create_statement)
    return ftis

def _gen_relationship_decs(model:_t.model_type)->Iterable[str]:
    '''Generates a list of foreign key statements for a table'''
    value = []
    rels = model.relationships
    if isinstance(rels,(list)):
        for r in rels:
            value.append(r.create_statement)
    return value

def _gen_table_create_contents(model:_t.model_type):
    '''
        Generates the contents of an SQL create table statement

        Columns,primary keys,foreign keys,unique constraints.

        ----------

        Arguments
        -------------------------
        `model` {model}
            The model instance to generate the contents for.


        Return {str}
        ----------------------
        The string SQL contents for the table.

        Meta
        ----------
        `author`: Colemen Atwood
        `created`: 03-25-2023 08:26:41
        `memberOf`: MySQLGenerator
        `version`: 1.0
        `method_name`: _gen_table_create_contents
        * @xxx [03-25-2023 08:28:07]: documentation for _gen_table_create_contents
    '''
    value = []
    value = c.arr.append(value,_gen_table_create_columns(model),skip_nulls=True)
    value = c.arr.append(value,_gen_primary_columns(model),skip_nulls=True)
    value = c.arr.append(value,_gen_unique_constraints(model),skip_nulls=True)
    value = c.arr.append(value,_gen_fulltext_indexes(model),skip_nulls=True)
    value = c.arr.append(value,_gen_relationship_decs(model),skip_nulls=True)

    value = c.arr.strip_list_nulls(value)

    return ',\n'.join(value)


_TABLE_CREATE_TEMPLATE = """
CREATE TABLE $not_exists$table_name
(
$contents
)$table_comment;"""

def _gen_table_create_statement(model:_t.model_type):
    exists = _gen_safe_script_not_exists()

    s = Template(_TABLE_CREATE_TEMPLATE)
    value = s.substitute(
        not_exists=exists,
        table_name=model.quoted_name,
        contents=_gen_table_create_contents(model),
        table_comment=_gen_table_comment(model.model_description),
    )
    
    if _settings.globe.flavor in ["sqlite"]:
        result = _gen_virtual_fulltexts(model)
        if result is not None:
            value = f"{value}\n\n{result}"
    
    return value


_SQLITE_VIRTUAL_TABLE_FULLTEXT = """CREATE VIRTUAL TABLE ${not_exists}${qc}$virtual_table_name${qc} USING fts5(
    $columns,
    content='$table_name',
    content_rowid='$table_primary_id'
    );
$create_trigger
$update_trigger
$delete_trigger"""

def _gen_virtual_fulltexts(model:_t.model_type):
    '''
        Generates the virtual table used for full text querying on an SQLite table.
        
        If the model does not have a column with a fulltext index, this will generate nothing.

        ----------

        Arguments
        -------------------------
        `model` {model}
            The model instance to operate on.

        Return {str}
        ----------------------
        The virtual table declaration if the model requires it, None otherwise.

        Meta
        ----------
        `author`: Colemen Atwood
        `created`: 05-08-2023 10:39:56
        `version`: 1.0
        `method_name`: _gen_virtual_fulltexts
        * @TODO []: documentation for _gen_virtual_fulltexts
    '''
    cols = model.columns
    fulltext_cols:Iterable[_t.column_type]= []
    col_names:Iterable[str]= []
    for col in cols:
        if col.fulltext is True:
            fulltext_cols.append(col)
            col_names.append(col.name)
    if len(fulltext_cols) == 0:
        return None

    quoted_cols = [x.quoted_name for x in fulltext_cols]
    columns = ', '.join(quoted_cols)
    virtual_table_name = f"{model.model_name}_fts"
    s = Template(_SQLITE_VIRTUAL_TABLE_FULLTEXT)
    value = s.substitute(
        qc=_settings.control.sql_quote_char,
        virtual_table_name=virtual_table_name,
        table_name=model.model_name,
        columns=columns,
        not_exists=_gen_safe_script_not_exists(),
        table_primary_id=model.primary_column.name,
        create_trigger=_fts_create_trigger(model,virtual_table_name,fulltext_cols),
        update_trigger=_fts_update_trigger(model,virtual_table_name,col_names),
        delete_trigger=_fts_delete_trigger(model,virtual_table_name),
    )
    return value

_FTS_CREATE_TRIGGER_TEMPLATE = """CREATE TRIGGER $not_exists$trigger_name AFTER INSERT ON $table_name BEGIN
    INSERT INTO $virtual_table_name ($quote_column_names) VALUES($new_columns);
END;"""
def _fts_create_trigger(model:_t.model_type,virtual_table_name:str,columns:Iterable[_t.column_type]):
    # columns.insert(0,model.primary_column)
    # columns.insert(0,model.primary_column)

    quote_column_names = ', '.join([x.name for x in columns])
    quote_column_names = f"rowid,{quote_column_names}"
    new_columns = ', '.join([f"new.{x.name}" for x in columns])
    new_columns = f"new.{model.primary_column.name},{new_columns}"
    trigger_name = f"after_insert_{model.model_name}_fts"
    s = Template(_FTS_CREATE_TRIGGER_TEMPLATE)
    value = s.substitute(
        quote_column_names=quote_column_names,
        not_exists=_gen_safe_script_not_exists(),
        new_columns=new_columns,
        table_name=model.model_name,
        trigger_name=trigger_name,
        virtual_table_name=virtual_table_name,
    )
    return value


_FTS_UPDATE_TRIGGER_TEMPLATE = """CREATE TRIGGER $not_exists$trigger_name AFTER UPDATE ON $table_name BEGIN
    UPDATE $virtual_table_name SET $set_statements WHERE rowid = old.$primary_column_name;
END;"""
def _fts_update_trigger(model:_t.model_type,virtual_table_name:str,col_names:Iterable[str]):
    col_names = c.arr.force_list(col_names)
    set_statements = ', '.join([f"{x} = new.{x}" for x in col_names])
    
    trigger_name = f"after_update_{model.model_name}_fts"
    s = Template(_FTS_UPDATE_TRIGGER_TEMPLATE)
    value = s.substitute(
        primary_column_name=model.primary_column.name,
        not_exists=_gen_safe_script_not_exists(),
        set_statements=set_statements,
        table_name=model.model_name,
        trigger_name=trigger_name,
        virtual_table_name=virtual_table_name,
    )
    return value
_FTS_DELETE_TRIGGER_TEMPLATE = """CREATE TRIGGER $not_exists$trigger_name AFTER DELETE ON $table_name BEGIN
    DELETE FROM virtual_table_name WHERE rowid = old.$primary_column_name;
END;"""
def _fts_delete_trigger(model:_t.model_type,virtual_table_name:str):
    
    trigger_name = f"after_delete_{model.model_name}_fts"
    s = Template(_FTS_DELETE_TRIGGER_TEMPLATE)
    value = s.substitute(
        primary_column_name=model.primary_column.name,
        not_exists=_gen_safe_script_not_exists(),
        table_name=model.model_name,
        trigger_name=trigger_name,
        virtual_table_name=virtual_table_name,
    )
    return value

# ---------------------------------------------------------------------------- #
#                              GENERAL GENERATORS                              #
# ---------------------------------------------------------------------------- #


def _gen_safe_script_exists():
    '''Generates "IF EXISTS" if _settings.control.mysql.safe_scripts is True'''
    exists = ""
    if _settings.control.mysql.safe_scripts:
        exists = "IF EXISTS "
    return exists

def _gen_safe_script_not_exists():
    '''Generates "IF NOT EXISTS" if _settings.control.mysql.safe_scripts is True'''
    exists = ""
    if _settings.control.mysql.safe_scripts:
        exists = "IF NOT EXISTS "
    return exists









# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=line-too-long
# pylint: disable=unused-import

from dataclasses import dataclass
from typing import Iterable, OrderedDict, Sequence, Sized, Any, TypeVar
from collections import OrderedDict
from abc import ABC, abstractmethod


import colemen_utils as c

import volent.settings as _settings
import volent.settings.types as _t
from volent.exceptions import ValidationError
from volent.validate.Validator import Validator


_T = TypeVar("_T")

class NoneOf(Validator):

    field_name = None
    default_message = "Invalid Input"


    def __init__(
        self,
        iterable=Iterable,
        error:str=None,
        ):
        '''
            A validator that passes if the value does not match any value in the iterable.
            ----------

            Arguments
            -------------------------
            `iterable` {list}
                A list of values that the test value cannot contain.

            `error` {str}
                The custom error to use if this validator fails.
                Accepts `input`, `iterable`, `field_name`

            Return {any}
            ----------------------
            The test value is successful, raises ValidationError if it fails.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 03-27-2023 09:10:41
            `memberOf`: NoneOf
            `version`: 1.0
            `method_name`: NoneOf
            * @xxx [03-27-2023 09:13:12]: documentation for NoneOf
        '''
        self.iterable = iterable
        self.iterable_str = ', '.join(str(choice) for choice in iterable)
        self.error = error or self.default_message  # type: str

    def _format_error(self, value: Any) -> str:
        args = {
            "input":value,
            "iterable":self.iterable_str,
        }
        if self.field_name is not None:
            args["field_name"] = self.field_name

        return self.error.format(**args)


    def _repr_args(self) -> str:
        return f"iterable={self.iterable!r}"


    def __call__(self,value:Any,field_name:str=None)->Any:
        self.field_name = field_name
        
        try:
            if value in self.iterable:
                raise ValidationError(self._format_error(value))
        except TypeError:
            pass
        test_value = c.arr.force_list(value)

        return value


# class ContainsOnly(OneOf):

#     field_name = None
#     default_message = "One or more of the choices you made was not in: {choices}."


#     def __init__(
#         self,
#         choices=Iterable,
#         ):

#         self.choices = choices
#         self.choices_str = ', '.join(str(choice) for choice in self.choices)

#     def _format_error(self, value: str, message: str) -> str:
#         if self.field_name is not None:
#             message = f"{self.field_name} was not in {self.choices}"
#         return message.format(
#             input=value, choices=self.choices_str
#         )


#     def _repr_args(self) -> str:
#         return f"min={self.min!r}, max={self.max!r}, equal={self.equal!r}"


#     def __call__(self,value:Sequence[_T],field_name:str=None)->Sequence[_T]:
#         self.field_name = field_name
#         test_value = c.arr.force_list(value)

#         for val in test_value:
#             if val not in self.choices:
#                 raise ValidationError(self._format_error(value,self.default_message))

#         return value



# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=line-too-long
# pylint: disable=unused-import


from dataclasses import dataclass
from typing import Iterable,OrderedDict, Union


import colemen_utils as c


import volent.settings.types as _t
import volent.settings as _settings
from volent.Column import Column as _column
from volent.Relationship import Relationship as _relationship
from volent.UniqueConstraint import UniqueConstraint as _uniqueConstraint
from volent.mixins import MySQLGeneratorMixin
from collections import OrderedDict
from volent.exceptions import ValidationError


@dataclass
class Field:
    main = None
    # database:_t.database_type = None
    model:_t.model_type = None
    schema:_t.schema_type = None


    name:str = None

    # _fields:Iterable[_t.column_type] = None
    _description:str = None
    _example:str = None
    _ignore_null:bool = False
    column:_t.column_type = None


    required:bool = False
    nullable:bool = True
    default = _settings.types.no_default
    validators = None
    _aliases = None
    _data_type:_t.type_base_type = None
    # _value = None
    _value = _t.undefined
    '''The value associated to this field if there is no column'''
    _open_api_location:str = None
    _not_provided:bool = False


    def __init__(
        self,
        data_type:_t.type_base_type=None,
        column:str=None,
        required:bool=False,
        nullable:bool=True,
        empty_string_is_null:bool=True,
        ignore_null:bool=False,
        default=_settings.types.no_default,
        validate=None,
        description:str=None,
        example=None,
        open_api_location=None,
        aliases:Union[list,str]=None,
        ):
        '''
            Create a schema Field
            ----------

            Arguments
            -------------------------
            [`column`=None] {str}
                The name of the column that this field represents.
                The dot delimited path to the column

                If None, it will attempt to find a matching column in the model.

            [`required`=False] {bool}
                True if this field MUST have a value.
                
            [`nullable`=True] {bool}
                False if this field does not allow null values.

            [`empty_string_is_null`=True] {bool}
                Treat empty strings as None

            [`default`] {any}
                The default value to assign to this field

            [`validate`=None] {any}
                A list of validators to apply to this field

            [`ignore_null`=False] {bool}
                A list of validators to apply to this field

            [`description`=None] {any}
                A description of what this form stores.

            [`example`=None] {any}
                An example value of what this field stores.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 03-26-2023 09:34:14
            `memberOf`: Field
            `version`: 1.0
            `method_name`: Field
            * @xxx [04-14-2023 08:24:58]: documentation for Field
        '''
        self.column = column
        self.default = default
        self._description = description
        self._example = example
        self.required = required
        self.nullable = nullable
        self.empty_string_is_null = empty_string_is_null
        self._ignore_null = ignore_null
        self._data_type = data_type
        self._open_api_location = open_api_location
        self.aliases = aliases
        

        self.validators = c.arr.force_list(validate,allow_nulls=False)


    @property
    def summary(self):
        '''
            Get this Model's summary

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 03-23-2023 14:39:08
            `@memberOf`: Model
            `@property`: summary
        '''
        value = {
            "name":self.name,
        }
        return value

    @property
    def dump_only(self):
        '''
            Get this Field's dump_only

            If True, this field can only be retrieved from the database and not inserted or updated.

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 04-18-2023 15:42:42
            `@memberOf`: Field
            `@property`: dump_only
        '''
        value = self.column.dump_only
        return value

    @property
    def value(self):
        '''
            Get this Field's value

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 03-25-2023 12:02:50
            `@memberOf`: Field
            `@property`: value
        '''
        value = None
        if self.column is not None:
            value = self.column.value
        else:
            if self._value != _t.undefined:
                value = self._value
            else:
                if self.default != _t.no_default:
                    value=self._value = self.default
                    # value = self.default

        if isinstance(value,(str)):
            if len(value)==0 and self.empty_string_is_null:
                value = None
        return value

    @value.setter
    def value(self,value):
        '''
            Set the Field's value property

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 04-24-2023 08:52:25
            `@memberOf`: Field
            `@property`: value
        '''
        if self.column is not None:
            self.column.value = value
        else:
            self._value = value

    @property
    def data_type_instance(self):
        '''
            Get this Field's data_type_instance

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 04-25-2023 14:11:58
            `@memberOf`: Field
            `@property`: data_type_instance
        '''
        if self._data_type is not None:
            return self._data_type
        elif self.column is not None:
            return self.column.data_type

    @property
    def data_type(self):
        '''
            Get this Field's data_type

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 04-24-2023 08:42:36
            `@memberOf`: Field
            `@property`: data_type
        '''
        value = None
        if self._data_type is not None:
            if hasattr(self._data_type,"open_api_data_type"):
                value = self._data_type.open_api_data_type
        elif self.column is not None:
            value = self.column.deserialized_value()
        return value

    @property
    def aliases(self):
        '''
            Get this Field's aliases

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 04-27-2023 12:58:05
            `@memberOf`: Field
            `@property`: aliases
        '''
        value = self._aliases
        if value is None:
            value = []
            self._aliases = value
        return value

    @aliases.setter
    def aliases(self,value):
        '''
            Set the Field's aliases property

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 04-27-2023 12:58:24
            `@memberOf`: Field
            `@property`: aliases
        '''
        if isinstance(value,(str)):
            self._aliases = c.arr.force_list(str)
        if isinstance(value,(list)):
            self._aliases = value
        if value is None:
            self._aliases = []


    def validate_value(self,value):
        val = value
        if self._data_type is not None:
            if hasattr(self._data_type,"serializer"):
                val = self._data_type.serializer(val,self.name)
            if hasattr(self._data_type,"deserialized_value"):
                val = self._data_type.deserialized_value(val)


        if isinstance(val,self._data_type.python_data_type) is False:
            if self.nullable is True and val is None:
                pass
            else:
                raise ValidationError(f"{self.name} expects {self._data_type.python_data_type} types, {type(value)} provided.",self.name)

        # if self._less_than_data_len(val) is False:
        #     raise ValidationError(f"{self.name} is too long.",self.name)



        if self._is_null(val):
            raise ValidationError(f"{self.name} cannot be null.",self.name)


        for valid in self.validators:
            val = valid(val,self.name)

    def validate(self):
        '''
            Executes the validation methods on this field's value.

            if this field is not associated to a column, it will automatically use the validate_value method.
            This is important because columns require additional steps before they can be validated.


            ----------

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 04-24-2023 09:50:01
            `memberOf`: Field
            `version`: 1.0
            `method_name`: validate
            * @xxx [04-24-2023 09:54:22]: documentation for validate
        '''
        if self.value is None and self._ignore_null is True:
            return
        if self.column is None:
            return self.validate_value(self.value)
        # val = self.value
        val = self.column.deserialized_value
        # val = self.column.data_type.

        if isinstance(val,self.column.data_type.python_data_type) is False:
            if self.column.is_primary is True:
                pass
            elif self.column.nullable is True:
                pass
            else:
                raise ValidationError(f"{self.column.name} expects {self.column.data_type.python_data_type} types, {type(self.value)} provided.",self.name)

        if self._less_than_data_len(val) is False:
            raise ValidationError(f"{self.column.name} is too long.",self.name)

        if self._is_null(val):
            raise ValidationError(f"{self.name} cannot be null.",self.name)


        for valid in self.validators:
            val = valid(val,self.name)

    def _is_null(self,val):
        if isinstance(val,(str)):
            if len(val) == 0:
                val = None

        if self.nullable is False and self.value is None:
            return True
        return False

    def _less_than_data_len(self,val):
        if isinstance(self.column.data_type.data_length,(int)):
            val = str(val)
            length = len(val)
            if length > self.column.data_type.data_length:
                return False
        return True

        # "request_id":{
        #     "type":"string",
        #     "description":"The encoded request id",
        #     "example":encoded_id()
        # },

    def open_api_definition(self):
        if self.column is not None and self._description is None:
            self._description = self.column.comment
        data = {
            "type":self.data_type,
            "description":self._description,
            "example":self._example,
        }
        return data

    def open_api_data(self,loc:str="body")->dict:
        '''
            Generate a dictionary representation of this field that can be used with the open api library.

            ----------

            Arguments
            -------------------------
            [`loc`='body'] {str}
                Where the field's value can be found in a request ['body','path']

            Return {dict}
            ----------------------
            A dictionary of data about this field.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 04-24-2023 09:22:02
            `memberOf`: Field
            `version`: 1.0
            `method_name`: open_api_data.
            `throws`: ValueError - if the loc is not "body"/"path"
            * @xxx [04-24-2023 09:23:07]: documentation for open_api_data
        '''
        if self._open_api_location is not None:
            loc = self._open_api_location
        if loc not in _settings.control.open_api_param_locations:
            raise ValueError(f"{loc} is not a valid param location, expected: {','.join(_settings.control.open_api_param_locations)}")
        if self.column is not None and self._description is None:
            self._description = self.column.comment

        data = {
            "name": self.name,
            "description":self._description,
            "example":self._example,
            "in": loc,
            "type": self.data_type,
            "required": self.required,
        }
        if self.required is False:
            if self.default != _settings.types.no_default:
                data['default'] = self.default
            # else:
                # print(f"setting {self.name} required to True")
                # data['required'] = True

        return data


    @property
    def should_ignore(self):
        if self.value is None and self._ignore_null is True:
            return True
        return False

    @property
    def not_provided(self)->bool:
        return self._not_provided

    @not_provided.setter
    def not_provided(self,value):
        '''
            Set the Field's not_provided property

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 04-27-2023 13:42:47
            `@memberOf`: Field
            `@property`: not_provided
        '''
        self._not_provided = value

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} : {self.name}>"


    def name_in_dict(self,data):
        if self.name in data:
            return self.name
        for k,_ in data.items():
            if k in self.aliases:
                return k
        return False












from volent.validate.Validator import Validator
from volent.validate.Length import Length
from volent.validate.OneOf import OneOf,ContainsOnly
from volent.validate.Email import Email
from volent.validate.Equal import Equal
from volent.validate.GreaterThan import GreaterThan
from volent.validate.LessThan import LessThan
from volent.validate.NoneOf import NoneOf
from volent.validate.Range import Range
from volent.validate.Regex import Regex
from volent.validate.IpAddress import IpAddress
from volent.validate.CreditCardNumber import CreditCardNumber
from volent.validate.PhoneNumber import PhoneNumber
from volent.validate.SocialSecurityNumber import SocialSecurityNumber
from volent.validate.StrongPassword import StrongPassword
from volent.validate.FutureUnixDate import FutureUnixDate
from volent.validate.PastUnixDate import PastUnixDate


# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=line-too-long
# pylint: disable=unused-import
'''
    The full text index module

    ----------

    Meta
    ----------
    `author`: Colemen Atwood
    `created`: 04-28-2023 06:53:30
    `name`: FullTextIndex
    * @xxx [04-28-2023 06:53:50]: documentation for FullTextIndex
'''




from dataclasses import dataclass
from typing import Iterable
import colemen_utils as c
# import volent.settings as _settings
import volent.settings.types as _t
from volent.mixins.MySQLGeneratorMixin import MySQLGeneratorMixin

@dataclass
class FullTextIndex(MySQLGeneratorMixin):
    '''The class used to define a fulltext index on a model.'''
    main:_t._main_type = None
    # database:_t.database_type = None
    model:_t.model_type = None
    name:str = None
    comment:str = None
    columns:Iterable[_t.column_type] = None
    # comment:str = None




    def __init__(
        self,
        columns:Iterable[_t.column_type],
        name:str=None,
        # comment:str=None,
        ):
        '''
            Create a full text index on a model's column(s).

            ----------

            Arguments
            -------------------------
            `Columns` {list,column}
                A column or list of columns to create a fulltext index on.

            `name` {str}
                The name of this fulltext index, this does not appear in the SQL but can be used
                within volent to find a specific fulltext index.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 04-28-2023 06:54:27
            `memberOf`: FullTextIndex
            `version`: 1.0
            `method_name`: FullTextIndex
            * @xxx [04-28-2023 06:56:30]: documentation for FullTextIndex
        '''
        self.name = name
        # self.comment = comment
        self.columns = c.arr.force_list(columns)



        # _parse_parent(self,parent)


    @property
    def summary(self):
        '''
            Get this FullTextIndex's summary

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 03-23-2023 14:53:39
            `@memberOf`: FullTextIndex
            `@property`: summary
        '''
        value = {
            "name":self.name,
            # "comment":self.comment,
            "columns":[x.name for x in self.columns],
        }
        return value


    def __hash__(self):
        col_names = [x.name for x in self.columns]
        val = col_names.append(self.name)
        return hash(val)

    def __eq__(self,other):
        if isinstance(other,FullTextIndex):
            return self.__hash__() == other.__hash__()
        return False




# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=line-too-long
# pylint: disable=unused-import

from dataclasses import dataclass
from typing import Iterable
import colemen_utils as c
import volent.settings as _settings
import volent.settings.types as _t
from volent.exceptions import ValidationError
from volent.data_types.TypeBase import TypeBase as _type_base

@dataclass
class Integer(_type_base):

    def __init__(self,data_length:int=None) -> None:
        super().__init__(data_length)
        self.min_data_length = 0
        self.max_data_length = 4294967295
        self.sql_type_name = "INTEGER"
        self.python_data_type = (int)
        self.open_api_data_type = "integer"

        self._validate_data_length(data_length)


    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} : {self.data_length}>"

    def serialized_value(self,value,field_name:str=None):
        if isinstance(value,(int)):
            return value

        if value is None or value == _t.undefined:
            return value

        if isinstance(value,(str)):
            if c.valid.numeric_only(value):
                value = int(value)
        if isinstance(value,(float)):
            value = int(value)

        if isinstance(value,(int)) is False:
            raise ValidationError(f"Failed to serialize {value} to integer.",field_name)

        return value

    def deserialized_value(self,value,field_name:str=None):
        if isinstance(value,(int)):
            return value

        if value is None or value == _t.undefined:
            return value

        if isinstance(value,(str)):
            if c.valid.numeric_only(value):
                return int(value)
        
        if isinstance(value,(int)) is False:
            raise ValidationError(f"Failed to deserialize {value} to integer.",field_name)



# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=line-too-long
# pylint: disable=unused-import
'''
    The query delete module.

    ----------

    Meta
    ----------
    `author`: Colemen Atwood
    `created`: 04-28-2023 07:24:17
    `name`: delete
    * @xxx [04-28-2023 07:24:33]: documentation for delete
'''



from dataclasses import dataclass
from typing import Iterable, Union


import colemen_utils as c


import volent.settings as _settings
import volent.settings.types as _t
# import volent.settings as _settings
# from volent.Field import Field as _field
# from volent.Relationship import Relationship as _relationship
# from volent.UniqueConstraint import UniqueConstraint as _uniqueConstraint


@dataclass
class WhereClause:
    Query:_t.query_type = None
    column_name:str = None
    value:any = None
    max_value:any = None
    comparison:str = None
    # comparison_operator:str = None
    is_or_parent:bool = False
    or_parent_id:str = None
    or_child_count:int = 0
    is_or_child:bool = False


    def __init__(self,Query:_t.query_type,column_name:str,value,comparison:str) -> None:
        '''
            Create a delete query instance.

            ----------

            Arguments
            -------------------------
            `model` {model}
                The model instance to execute a delete operation on.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 04-28-2023 07:24:53
            `memberOf`: Delete
            `version`: 1.0
            `method_name`: Delete
            * @xxx [04-28-2023 07:26:55]: documentation for Delete
        '''
        self.Query = Query
        self._params = {}
        self.column_name = column_name
        self.comparison = comparison

        self.value = value or "NULL"
        snake_comp = c.string.to_snake_case(comparison)


        if snake_comp == "or":
            self.comparison = "or"
            self.is_or_parent = True
            self.or_parent_id = c.rand.css_class()


        if snake_comp in ["!","!=","isnt","isnot","is_not","<>"]:
            self.comparison = "is not"

        # if snake_comp in ["greater_than"]:
        #     self.comparison = ">"

        # if snake_comp in ["greater_than_equal"]:
        #     self.comparison = ">="

        if snake_comp in ["in","not_in"]:
            self.value = c.arr.force_list(self.value)


        if snake_comp == "between":
            self.value = value[0]
            self.max_value = value[1]

        if snake_comp == "in":
            self.value = c.arr.force_list(value)


    def __repr__(self) -> str:
        if self.is_or_parent:
            return f"<WhereClause OR: {self.or_parent_id} :: {self.or_child_count}>"
        return f"<WhereClause : {self.column_name} :: {self.comparison} :: {self.value} :: {self.or_parent_id}>"

    def set_or_id(self,or_id):
        print(self)
        print(f"set or id: {or_id}")
        self.or_parent_id = or_id
        print(self)

    @property
    def comparison_operator(self):
        '''
            Get this WhereClause's comparison_operator

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 04-01-2024 16:18:47
            `@memberOf`: WhereClause
            `@property`: comparison_operator
        '''

        snake_comp = c.string.to_snake_case(self.comparison)
        match snake_comp:
            case "greater_than":
                return ">"
            case "greater_than_equal":
                return ">="
            case "less_than":
                return "<"
            case "less_than_equal":
                return "<="
        if self.comparison in [">","<","=","!=",">=","<="]:
            return self.comparison
        return self.comparison.upper()



    def where_string(self):
        '''Generate the SQL for this where clause'''
        snake_comp = c.string.to_snake_case(self.comparison)
        # if self.is_or_parent is True:

        if snake_comp in ["or"]:
            subwheres = []
            children = self.Query.get_wheres_by_or_id(self.or_parent_id)
            # print(f"whereClause.where_string : or child count : {len(children)}")
            for ch in children:
                cws = ch.where_string()
                # print(f"cws: {cws}")
                subwheres.append(cws)
            sub_string = ' OR '.join(subwheres)
            if self.Query.where_count > 0:
                sub_string = f"({sub_string})"
            return sub_string



        if snake_comp in ["is_not"]:
            if self.value != "NULL":
                return f"{self.column_name} != {self.quoted_value}"


        if snake_comp in ["is","is_not"]:
            comp = self.comparison.upper()
            p = self.Query.add_param(c.rand.css_class(),self.value)
            if self.should_quote:
                comp = "!="
                if snake_comp == "is":
                    comp = "="
            return f"{self.column_name} {comp} {p.parameterized_name}"

        if snake_comp in ["greater_than","greater_than_equal","less_than","less_than_equal"]:
            p = self.Query.add_param(c.rand.css_class(),self.value)
            return f"{self.column_name} {self.comparison_operator} {p.parameterized_name}"

        if snake_comp in ["between"]:
            min_key = c.rand.rand()
            max_key = c.rand.rand()
            min_param = self.Query.add_param(min_key,self.value)
            max_param = self.Query.add_param(max_key,self.max_value)
            # min_key =f"{wcol}_minimum"
            # max_key =f"{wcol}_maximum"

            return f"{self.column_name} {self.comparison.upper()} {min_param.parameterized_name} AND {max_param.parameterized_name}"


        if snake_comp in ["between"]:
            return f"{self.column_name} {self.comparison.upper()} {self.quoted_value}"


        if snake_comp in ["in","not_in"]:
            inlist = []
            for v in self.value:
                key = c.rand.rand()
                if _settings.globe.flavor in ['sqlite']:
                    inlist.append(f"@{key}")
                    self.Query.add_param(key,v)
            inlist_string = ','.join(inlist)
            return f"{self.column_name} {self.comparison.upper()} ({inlist_string})"


        if snake_comp in ["like"]:
            # p = self.Query.add_param(c.rand.css_class(),self.value)
            # return f"{self.column_name} {self.comparison.upper()} '{p.parameterized_name}'"
            return f"{self.column_name} {self.comparison.upper()} '{self.value}'"



    @property
    def quoted_value(self):
        '''
            returns the value wrapped in quotes IF the column's python data type is string.
        '''
        if self.should_quote:
            return f"'{self.value}'"
        return self.value


    @property
    def should_quote(self):
        '''
            True if the column's python data type is a string.
        '''
        col = self.Query.model.get_column(self.column_name)
        # print(f"col.data_type : {col.data_type}")
        if self.value == "NULL":
            return False
        if isinstance(self.value,col.data_type.python_data_type):
            return True
        return False










# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=line-too-long
# pylint: disable=unused-import


from dataclasses import dataclass
import os
from typing import Iterable


import colemen_utils as c


import volent.settings.types as _t
import volent.settings as _settings
# import volent.settings as _settings
# from volent.Column import Column as _column
# from volent.Relationship import Relationship as _relationship
# from volent.UniqueConstraint import UniqueConstraint as _uniqueConstraint
# from volent.mixins import OrderedClass,MySQLGeneratorMixin
# from volent.query.Insert import Insert as _insert
from volent.mixins.DatabaseConnection import DatabaseConnection as _DatabaseConnection


# class Model(MySQLGeneratorMixin):
@dataclass
class Database(_DatabaseConnection):
    main:_t._main_type = None
    '''A reference to volent instance.'''

    _name:str = None
    '''The name of this database.'''

    _models:Iterable[_t.model_type] = None
    '''A list of model instances that belong to this database.'''


    _models_lookup:dict = None
    '''A dictionary of models used for lookups.'''

    def __init__(self,main:_t._main_type,name:str) -> None:
        self.main = main
        self._name = name
        self._models = []
        self._models_lookup = {}

    def generate_sql(self,file_path:str=None,drops:bool=True)->str:
        '''
            Generates the master sql for all tables in this database.

            ----------

            Arguments
            -------------------------
            [`file_path`=None] {str}
                The path to where the sql will be saved, defaults to {cwd}/master.sql
            [`drops`=True] {bool}
                If False, the drop statements will not be added to the sql.


            Return {str}
            ----------------------
            The sql content.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 04-28-2023 09:41:21
            `memberOf`: Database
            `version`: 1.0
            `method_name`: generate_sql
            * @xxx [04-28-2023 09:42:56]: documentation for generate_sql
        '''
        if file_path is None:
            file_path = f"{os.getcwd()}/{self.name}.sql"
        sql = self.master_sql(file_path,drops)
        return sql

    @property
    def summary(self):
        '''
            Get this Database's summary

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 04-18-2023 08:24:31
            `@memberOf`: Database
            `@property`: summary
        '''
        value = {
            "name":self.name,
            "models":[x.summary for x in self.models]
        }

        return value

    @property
    def name(self)->str:
        '''
            Get this Database's name

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 04-11-2023 14:15:56
            `@memberOf`: Database
            `@property`: name
        '''
        return self._name

    @property
    def models(self):
        '''
            Get this Database's models

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 04-18-2023 08:15:08
            `@memberOf`: Database
            `@property`: models
        '''
        value = self._models
        return value

    def get_model(self,name:str)->_t.model_type:
        '''
            Retrieve a model by searching its name.

            ----------

            Arguments
            -------------------------
            `name` {str}
                The name of the model to search for.

            Return {model}
            ----------------------
            The model instance if successful, None otherwise.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 04-28-2023 09:57:12
            `memberOf`: Database
            `version`: 1.0
            `method_name`: get_model
            * @xxx [04-28-2023 09:57:59]: documentation for get_model
        '''
        model = None
        if name in self._models_lookup:
            return self._models_lookup[name]
        for m in self.models:
            if m.name == name:
                return m

        return model


    def register_models(self):
        '''
            Iterate all registered model instances and associate any that have this database's name.

            This should ONLY be called by volent._generate_databases.
            Otherwise, leave this alone.

            ----------

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 04-18-2023 08:12:17
            `memberOf`: Database
            `version`: 1.0
            `method_name`: register_models
            * @xxx [04-18-2023 08:13:08]: documentation for register_models
        '''
        for mdl in self.main.models:
            if mdl.database_name == self.name:
                # print(f"associating model {mdl.name} to database {self.name}")
                mdl._database = self
                self._models.append(mdl)
                self._models_lookup[mdl.model_name] = mdl

    def existing_tables(self):
        sql = f"""SELECT * FROM information_schema.tables WHERE table_type = 'base table' AND TABLE_SCHEMA = '{self.name}'"""
        # print(f"sql:{sql}")
        self.run(sql)
        result = self.fetchall()
        # print(f"result:{result}")
        c.file.writer.to_json("live_db.json",result)
        for x in result:
            # if x['TABLE_SCHEMA'] == self.name:
            table_name = x['TABLE_NAME']
            if table_name not in self._models_lookup:
                c.con.log(f"No Model exists for table {table_name}","magenta")
            else:
                c.con.log(f"Model located for table {table_name}","blue")



            # table = self.get_model(table_name)
            # if table is None:
            #     c.con.log(f"model {table_name} does not exist in the database","magenta")
            # print(x['table_name'])



    def drop_tables(self):
        '''
            Drop all tables that are associated to this database.

            This will remove tables regardless of if a model exists, ALL TABLES WILL BE DROPPED.

            ----------

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 04-18-2023 09:58:49
            `memberOf`: Database
            `version`: 1.0
            `method_name`: drop_tables
            * @xxx [04-18-2023 10:15:24]: documentation for drop_tables
        '''
        sql = f"""SELECT table_name FROM information_schema.tables WHERE table_type = 'base table' AND TABLE_SCHEMA = '{self.name}'"""
        self.run(sql)
        result = self.fetchall()
        for x in result:
            name = x['table_name']
            # drop_statement = f"DROP TABLE `{self.name}`.`{name}`;"
            drop_statement = self.gen_drop_table_statement(name,self.name)
            print(f"        drop_statement:{drop_statement}")
            result = self.run(drop_statement,foreign_key_checks=False)
            if result is True:
                c.con.log(f"Successfully Dropped table: {name}","success")

    def drop_model_tables(self):
        '''
            Drop all tables that have an associated model.

            This means any tables that already exists in the database but do not have a model will be kept.

            ----------


            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 04-18-2023 10:19:45
            `memberOf`: Database
            `version`: 1.0
            `method_name`: drop_model_tables
            * @xxx [04-18-2023 10:20:36]: documentation for drop_model_tables
        '''
        for m in self.models:
            print(f"        Dropping model table: {m.model_name}")
            name = m.model_name
            # drop_statement = f"DROP TABLE `{self.name}`.`{name}`;"
            drop_statement = self.gen_drop_table_statement(name,self.name)
            print(f"        statement: {drop_statement}")
            result = self.run(drop_statement,foreign_key_checks=False)
            print(f"        result:{result}")
            if result is True:
                c.con.log(f"Successfully Dropped table: {name}","success")
            else:
                c.con.log(f"Failed to Drop table: {name}","error")

    def create_model_tables(self,force:bool=False,foreign_key_checks:bool = False):
        '''
            Create tables for all models associated to this database.

            ----------

            Arguments
            -------------------------
            [`force`=bool] {False}
                If True, this will drop the tables then recreate them

            [`foreign_key_checks`=False] {bool}
                If False, the database will be instructed to skip foreign_key_checks.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 04-18-2023 10:17:07
            `memberOf`: Database
            `version`: 1.0
            `method_name`: create_model_tables
            * @xxx [04-18-2023 10:18:44]: documentation for create_model_tables
        '''
        for m in self.models:
            if force is True:
                self.run(m.drop_statement,foreign_key_checks=foreign_key_checks)
            self.run(m.create_statement,foreign_key_checks=foreign_key_checks)

    def last_id(self)->int:
        '''
            Retrieve the last insert id committed to the database.
            ----------

            Return {int}
            ----------------------
            The primary id of the last inserted row.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 12-16-2022 08:43:06
            `memberOf`: MySQLDatabase
            `version`: 1.0
            `method_name`: last_id
            * @xxx [12-16-2022 08:44:37]: documentation for last_id
        '''
        sql = 'SELECT LAST_INSERT_ID();'
        if _settings.globe.flavor in ['sqlite']:
            sql = "SELECT last_insert_rowid()"
        result = self.run_select(sql)
        if isinstance(result,(list)):
            try:
                result = result[0]['LAST_INSERT_ID()']
            except TypeError:
                result = result[0][0]
            except IndexError:
                result = result[0][0]
        return result












# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=line-too-long
# pylint: disable=unused-import

from dataclasses import dataclass
from typing import Iterable, OrderedDict, Sized, Union
from collections import OrderedDict
from abc import ABC, abstractmethod


import colemen_utils as c

import volent.settings as _settings
import volent.settings.types as _t
from volent.exceptions import ValidationError
from volent.validate.Validator import Validator

class Length(Validator):

    field_name = None
    message_min = "Shorter than minimum length {min}."
    message_max = "Longer than maximum length {max}."
    message_all = "Length must be between {min} and {max}."
    message_equal = "Length must be {equal}."


    def __init__(
        self,
        min=None,
        max=None,
        equal=None
        ):
        if equal is not None and any([min, max]):
            raise ValueError(
                "The `equal` parameter was provided, maximum or "
                "minimum parameter must not be provided."
            )

        self.min = min
        self.max = max
        self.equal = equal

    def _format_error(self, value: Sized, message: str) -> str:
        if self.field_name is not None:
            message = message[0].lower() + message[1:]
            message = f"{self.field_name} {message}"
        return message.format(
            input=value, min=self.min, max=self.max, equal=self.equal
        )


    def _repr_args(self) -> str:
        return f"min={self.min!r}, max={self.max!r}, equal={self.equal!r}"


    def __call__(self,value:Sized,field_name:str=None)->Sized:
        length = len(value)
        self.field_name = field_name

        if self.equal is not None:
            if length != self.equal:
                raise ValidationError(self._format_error(value,self.message_equal),field_name)
            return value

        if self.min is not None and length < self.min:
            message = self.message_min if self.max is None else self.message_all
            raise ValidationError(self._format_error(value, message),field_name)

        if self.max is not None and length > self.max:
            message = self.message_max if self.min is None else self.message_all
            raise ValidationError(self._format_error(value, message),field_name)
        return value


# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=line-too-long
# pylint: disable=unused-import


from dataclasses import dataclass
from typing import Iterable,OrderedDict, Union


import colemen_utils as c


import volent.settings.types as _t
import volent.settings as _settings
from volent.Field import Field as _field
from volent.NestedField import NestedField as _nested_field
from volent.mixins import OrderedClass
from volent.exceptions import ValidationError





@dataclass
class Schema(metaclass=OrderedClass):

    main = None
    # database:_t.database_type = None
    model:_t.model_type = None
    _table_name:str = None
    _name:str = None
    _fields:Iterable[_t.field_type] = None
    _nested_fields:Iterable[_t.nested_field_type] = None
    _schema_description:str = None
    _schema_crud_type:str = None

    __unique_prop_keys = None

    has_nested_schemas:bool = False

    # _field_aliases = None



    def __init__(
        self,
        model:_t.model_type=None
        ) -> None:
        from volent.Model import Model as _model
        if isinstance(model,_model):
            self.model = model
        else:
            if isinstance(self.table_name,(str)):
                mdl = _settings.globe.Volent.get_model(self.table_name)
                if mdl is not None:
                    self.model = mdl

            elif model is not None:
                md = model()

    #     self.from_dict(kwargs)


    @property
    def summary(self):
        '''
            Get this Schema's summary

            {
                name:str,
                fields:[]
            }

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 03-23-2023 14:39:08
            `@memberOf`: Schema
            `@property`: summary
        '''
        value = {
            "name":self.name.snake_,
            "fields": [x.summary for x in self.fields],
        }
        return value

    @property
    def table_name(self):
        '''
            Get this Schema's table_name

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 04-02-2024 12:30:25
            `@memberOf`: Schema
            `@property`: table_name
        '''
        value = self._table_name
        if value is None:
            value = None
            keys = ["__table_name__","__tablename__"]
            for key in keys:
                if hasattr(self,key):
                    value = getattr(self,key)
                    break
            self._table_name = value
        return value

    @property
    def schema_name(self)->str:
        '''
            Get this Schema's name

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 03-23-2023 14:02:44
            `@memberOf`: Schema
            `@property`: name
        '''
        # print(self.__dict__[])
        # print(f"======================================== SCHEMA")

        # print(self.__class__.__name__)
        value = self._name
        if value is None:
            self.custom_name_key = "__schema_name__"
            if hasattr(self,self.custom_name_key):
                value = getattr(self,self.custom_name_key)
            else:
                value = c.string.to_snake_case(self.__class__.__name__)

            # value = validate_table_name(value)
            # value = c.string.Name(value)
            self._name = value
            # if hasattr(self,"__model_name__"):
            #     value = self.__model_name__
            #     value = _settings.control.mysql.validate_table_name(value)
            #     value = c.string.Name(value)
            #     self._name = value
        return value

    @property
    def schema_description(self):
        '''
            Get this Schema's schema_description

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 03-24-2023 07:27:06
            `@memberOf`: Schema
            `@property`: schema_description
        '''
        # print(self.__class__.__doc__)
        value = self._schema_description
        if value is None:
            if hasattr(self,"__description__"):
                value = self.__description__
                # value = validate_table_comment(self.name.snake_,value)
                self._schema_description = value
            else:
                value = self.__class__.__doc__
        return value

    @property
    def schema_crud_type(self):
        '''
            Get this Schema's schema_crud_type

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 03-24-2023 07:27:06
            `@memberOf`: Schema
            `@property`: schema_crud_type
        '''
        value = self._schema_crud_type
        if value is None:
            if hasattr(self,"__schema_crud_type__"):
                value = self.__schema_crud_type__
                self._schema_crud_type = value
        return value

    @property
    def nested_fields(self):
        '''
            Get this Schema's nested_fields

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 05-07-2023 15:53:24
            `@memberOf`: Schema
            `@property`: nested_fields
        '''
        value = self._nested_fields
        if value is None:
            value = []
        return value

    @property
    def fields(self)->Iterable[_t.field_type]:
        '''
            Retrieve a list of field instances associated to this schema.

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 04-11-2023 14:07:35
            `@memberOf`: Model
            `@property`: fields
        '''
        value = self._fields
        nested_fields = []

        if value is None:
            dif = self._unique_prop_keys
            # df_props = dir(Model(_is_root=False))
            # # # @Mstep [] gather the props of this instance.
            # props = dir(self)
            # # # @Mstep [] find the props that exist on this instance and not on the base.
            # dif = c.arr.find_list_diff(props,df_props)
            # dif = dir(self)
            value = []
            nested_fields = []
            # color = c.rand.option(["magenta","yellow","green","cyan"])
            for prop in dif:
                name = prop
                # if isinstance(val,(dict)):
                    # props[name] = val
                val = getattr(self,prop)
                if isinstance(val,_nested_field):
                    val.name = c.string.to_snake_case(name)
                    val.schema = self
                    nested_fields.append(val)
                #     val.child_schema.fields()
                if isinstance(val,(_field)):
                    # c.con.log(f"located Field: {name}","green")
                    val.name = c.string.to_snake_case(name)
                    val.schema = self
                    value.append(val)



        order_cols = []
        for k in list(self.ordered_attrs(True)):
            for col in value:
                if col.name == k[0]:
                    order_cols.append(col)
                    break
        value = order_cols
        self._fields = value
        self._nested_fields = nested_fields
        return value


    def __call_custom_validations(self):
        '''
            Searches for methods that start with "valid" and executes them.

            This is allows custom validation methods to be added to the schema.
            They are executed before the standard validations so that the custom ones can also
            coerce values.

            If ANY exception is raised in the custom method, this will raise a validationError.

            ----------


            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 04-24-2023 09:41:34
            `memberOf`: Schema
            `version`: 1.0
            `method_name`: __call_custom_validations
            * @xxx [04-24-2023 09:44:30]: documentation for __call_custom_validations
        '''
        for func in dir(self):
            if func.startswith("valid"):
                vm = getattr(self,func)
                if callable(vm):
                    try:
                        vm()
                    except Exception as e:
                        raise ValidationError(e)

    def __validate(self):
        '''
            Execute the validation process on all fields associated to this schema

            This will execute the custom validations before calling built it ones.

            ----------

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 04-21-2023 07:43:15
            `memberOf`: Schema
            `version`: 1.0
            `method_name`: __validate
            * @xxx [04-21-2023 07:44:01]: documentation for __validate
        '''
        self.__call_custom_validations()
        for f in self.fields:
            f.validate()

    def __validate_dict(self,data:dict):
        '''
            Execute the validation process on all fields associated to this schema

            This is used to validate a dictionary as opposed to __validate which will
            validate the values that are assigned to columns.

            So this method has a little setup before it can validate which includes
            it associating dictionary keys to matching fields.


            ----------

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 04-21-2023 07:43:15
            `memberOf`: Schema
            `version`: 1.0
            `method_name`: __validate
            * @xxx [04-21-2023 07:44:01]: documentation for __validate
        '''
        # @Mstep [] convert all keys to snake case for matching.
        data = c.obj.keys_to_snake_case(data)
        # print(f"data:{data}")
        # @Mstep [LOOP] iterate all fields associated to this schema
        for f in self.fields:
            # @Mstep [] set "value" to be no_default
            value = _settings.types.no_default
            # @Mstep [IF] if field name is not in the data dict
            if f.name not in data:
                # @Mstep [IF] if the field is required.
                if f.required is True:
                    # @Mstep [] raise a validation error.
                    raise ValidationError(f"{f.name} is required.")
                # @Mstep [IF] if the field name converted to snake case is in the data dict.
                elif c.string.to_snake_case(f.name) in data:
                    # @Mstep [] update "value" to be the data dict's value.
                    value = data[c.string.to_snake_case(f.name)]
                # @Mstep [ELSE] the field was not provided
                else:
                    # @Mstep [] set the field._ignore_null to True, so this field will not attempt validation.
                    f._ignore_null = True
                    continue
            # @Mstep [ELSE] the field was found in the data dictionary.
            else:
                # @Mstep [] update "value" to be the data dict's value.
                value = data[f.name]

            # @Mstep [IF] if "value" is NOT no_default
            if value is not _settings.types.no_default:
                # @Mstep [] set the field's value to the "value"'s value.
                f.value = value

        # @Mstep [] Execute any custom validation methods on this schema
        self.__call_custom_validations()
        # @Mstep [LOOP] iterate the fields associated to this schema
        for f in self.fields:
            if f.not_provided is True:
                continue
            # @Mstep [IF] if the field should NOT be skipped
            if f.should_ignore is False:
                # @Mstep [] have the field self-validate
                f.validate()

    def get_field(self,name:str)->_t.field_type:
        '''
            Retrieve a field instance from this schema by its name.

            ----------

            Arguments
            -------------------------
            `name` {string}
                The name of the field to search for.

            Return {field}
            ----------------------
            The field instance if successful, None otherwise.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 04-21-2023 07:44:07
            `memberOf`: Schema
            `version`: 1.0
            `method_name`: get_field
            * @xxx [04-21-2023 07:44:52]: documentation for get_field
        '''
        # print(f"Schema.get_field.name : {name}")
        name = c.string.to_snake_case(name)
        for field in self.fields:
            if field.name == name:
                return field
        return None

    def from_dict(self,data:dict):
        '''
            Populate a model's column values from a dictionary.

            The keys must match the column names exactly.
            ----------

            Arguments
            -------------------------
            `data` {dict}
                A dictionary to populate the columns with.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 04-21-2023 07:45:36
            `memberOf`: Schema
            `version`: 1.0
            `method_name`: from_dict
            * @xxx [04-21-2023 07:48:45]: documentation for from_dict
        '''

        if isinstance(data,(dict)) is False:
            if c.types.is_list_of_dicts(data):
                self.from_list_of_dicts(data)
                return
            else:
                raise TypeError(f"from_dict expects a dictionary, {type(data)} received.")

        for col in self.model.columns:
            if col.name not in data:
                continue
            field = self.get_field(col.name)
            if field is None:
                continue

            # col.column_value = data[col.name]
            col.value = data[col.name]
        return self

    def from_list_of_dicts(self,data:Iterable[dict])->_t.schema_type:
        if c.types.is_list_of_dicts(data) is False:
            raise TypeError(f"from_list_of_dicts expects a list of dictionaries, {type(data)} received.")
        nm = self.__dicts_to_models(data)
        if nm is None:
            model = model if nm is None else nm

    def open_api_definition(self):
        '''
            Generate the open API definition for this schema.

            ----------

            Return {dict}
            ----------------------
            A dictionary definition of this schema.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 05-07-2023 15:30:49
            `memberOf`: Schema
            `version`: 1.0
            `method_name`: open_api_definition
            * @xxx [05-07-2023 15:32:48]: documentation for open_api_definition
        '''
        value = {}
        for f in self.fields:
            value[f.name] = f.open_api_definition()
        for f in self.nested_fields:
            value[f.name] = f.open_api_definition()

        return value

    def open_api_data(self,loc:str="body")->Iterable[dict]:
        '''
            Get this Schema's open_api_data

            returns a list of dictionaries each representing a field, the dictionaries can
            be used with the open api library for documentation.

            ----------

            Arguments
            -------------------------
            [`loc`='body'] {str}
                Where the field's value can be found in a request ['body','path']


            Return {list}
            ----------------------
            A list of dictionaries.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 04-24-2023 09:17:49
            `memberOf`: Schema
            `version`: 1.0
            `method_name`: open_api_data
            `throws`: ValueError - if the loc is not "body"/"path"
            * @xxx [04-24-2023 09:19:05]: documentation for open_api_data
        '''

        value = []
        if loc not in _settings.control.open_api_param_locations:
            raise ValueError(f"{loc} is not a valid param location, expected: {','.join(_settings.control.open_api_param_locations)}")
        for f in self.fields:
            value.append(f.open_api_data(loc))
        return value

    # def serialized(self,model:Union[list,dict,_t.model_type]=None):


    def __deserialize_dictionary(
        self,
        data:dict,
        ignore_unprovided:bool=False,
        )->dict:
        # c.con.log(f"Deserializing dictionary - raw: {data}","magenta")
        data = self._remove_missing_fields(data)
        # data = self._correlate_to_dict(data)
        # c.con.log(f"Deserializing dictionary - correlated: {data}","info")
        self.__validate_dict(data)
        out_data = {}
        for f in self.fields:
            if hasattr(f.data_type_instance,"deserialized_value"):
                if f.name in data:
                    out_data[f.name] = f.data_type_instance.deserialized_value(data[f.name],f.name)
                continue
        # c.con.log(f"Deserializing dictionary - out_data: {out_data}","green")
        return out_data

    def __deserialize_model(self,model:_t.model_type):
        out_data = {}

        self.model = model
        model = self._correlate_to_columns()
        self.__validate()
        # data = {}
        for f in self.fields:
            # if hasattr(f.data_type_instance,"serializer"):
            #     v = f.data_type_instance.serializer(f.value,f.name)
            #     out_data[f.name] = v
            #     continue
            if hasattr(f.data_type_instance,"deserialized_value"):
                v = f.data_type_instance.deserialized_value(f.value)
                out_data[f.name] = v
                continue
            v = f.value
            out_data[f.name] = v

        return out_data

    def deserialized(
        self,
        model:Union[list,dict,_t.model_type]=None,
        many:bool=False,
        ignore_unprovided=False
        ):
        from volent.Model import Model as _model

        if isinstance(model,(dict)) is True:
            dr = self.__deserialize_dictionary(model,ignore_unprovided)
            if many:
                dr = c.arr.force_list(dr)
            return dr

        if isinstance(model,(list)) is True:
            out_data = []
            for mdl in model:
                data = {}
                if isinstance(mdl,dict):
                    data = self.__deserialize_dictionary(mdl)
                    out_data.append(data)
                    continue

                if isinstance(mdl,_model):
                    data = self.__deserialize_model(mdl)
                    out_data.append(data)
                    continue

                # out_data.append(data)

            if many:
                out_data = c.arr.force_list(out_data)
            return out_data

        if isinstance(model,(_model)) is True:
            out_data = self.__deserialize_model(mdl)
            if many:
                out_data = c.arr.force_list(out_data)
            return out_data
    load = deserialized

    def __dicts_to_models(self,data:Iterable[dict])->Union[Iterable[_t.model_type],None]:
        '''
            This will attempt to convert a list of dictionaries into a list of model instances.

            Arguments
            -------------------------
            `data` {list}
                A list of dictionaries to convert

            Return {list}
            ----------------------
            A list of models if successful, None otherwise

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 04-02-2024 13:28:23
            `memberOf`: Schema
            `version`: 1.0
            `method_name`: __dicts_to_models
            * @xxx [04-02-2024 13:29:51]: documentation for __dicts_to_models
        '''
        from volent.Model import Model as _model
        nm = []
        if c.types.is_list_of_dicts(data):
            if isinstance(self.model,_model) is True:
                for mdl in data:
                    nc = self.model.__class__
                    n = nc(**mdl)
                    nm.append(n)
        if len(nm) == 0:
            return None
        return nm

    def dump(self,model:_t.model_type=None,many=False,ignore_unprovided=False)->Union[dict,Iterable[dict]]:
        '''Dump the contents of the model(s) using the fields to filter.'''

        if isinstance(model,(dict)):
            # print(f"dictionary provided for dumping.")
            data = self._correlate_to_dict(model,ignore_unprovided)
            self.__validate_dict(model)
            out_data = {}
            for f in self.fields:
                if f.name in data:
                    # print(f"f.value:{f.value}")
                    if hasattr(f.data_type_instance,"serializer"):
                        v = f.data_type_instance.serializer(data[f.name],f.name)
                        out_data[f.name] = v
                        continue
                    # if hasattr(f.data_type_instance,"serialized_value"):
                    #     v = f.data_type_instance.serialized_value(data[f.name])
                    #     print(f"{f.name}:: {v}")
                    #     out_data[f.name] = v
                    #     continue
                    out_data[f.name] = f.value
                    # out_data[f.name] = data[f.name]
            if many:
                out_data = c.arr.force_list(out_data)
            return out_data




        # print(f"model: {model}")
        if isinstance(model,(list)):
            # print(f"a list was provided for dumping.")
            result = []
            from volent.Model import Model as _model
            nm = self.__dicts_to_models(model)
            model = model if nm is None else nm
            # nm = []
            # if c.types.is_list_of_dicts(model):
            #     if isinstance(self.model,_model) is True:
            #         for mdl in model:
            #             nc = self.model.__class__
            #             n = nc(**mdl)
            #             nm.append(n)
            #         model = nm

            for mdl in model:

                if isinstance(mdl,_model) is False:
                    c.con.log(f"The mdl is not an instance of model:{mdl}","magenta")


                # mdl = model

                self.model = mdl
                self._correlate_to_columns()
                self.__validate()
                data = {}
                for f in self.fields:
                    if hasattr(f.data_type_instance,"serializer"):
                        v = f.data_type_instance.serializer(f.value,f.name)
                        data[f.name] = v
                        continue
                    if hasattr(f.data_type_instance,"serialized_value"):
                        v = f.data_type_instance.serialized_value(f.value)
                        # print(f"{f.name}:: {v}")
                        # print(f"v: {v}")
                        data[f.name] = v
                        continue
                    v = f.value
                    data[f.name] = v
                    # data[f.name] = f.value
                result.append(data)
            return result


        if model is not None:
            self.model = model
        # self.model = model
        self._correlate_to_columns()
        self.__validate()
        data = {}
        for f in self.fields:
            # print(f"{f.name}:")
            if f.should_ignore is False:
                if hasattr(f.data_type_instance,"serializer"):
                    v = f.data_type_instance.serializer(f.value,f.name)
                    data[f.name] = v
                    continue
                if hasattr(f.data_type_instance,"serialized_value"):
                    v = f.data_type_instance.serialized_value(f.value)
                    # print(f"{f.name}:: {v}")
                    data[f.name] = v
                    continue
                data[f.name] = f.value
        if many:
            data = c.arr.force_list(data)
        return data

    def new(self,data:dict=None)->_t.schema_type:
        '''
            Validate this schema's data and submit it to the model's table in the databse.


            Arguments
            ----------------------
            `data` {dict,None}
                A dictionary to populate the columns with.

                This is the same as calling "schema.from_dict(data).new()

            Return {schema}
            ----------------------
            This schema instance.

            Example
            ---
            ```
                songData = {
                    "title":f"{c.rand.abstract_name()}{c.rand.number()}",
                    "artist":c.rand.abstract_name(),
                    "lyrics":"bobs_borbers on my norblers.",
                }

                SongsNewSchema().new(songData)
            ```




            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 04-21-2023 07:50:29
            `memberOf`: Schema
            `version`: 1.0
            `method_name`: new
            * @xxx [04-21-2023 07:51:43]: documentation for new
        '''
        if data is not None:
            if isinstance(data,(dict)) is False:
                raise TypeError(f"Schema.new() expects the data argument to be a dictionary, {type(data).__name__} provided.")
            else:
                self.from_dict(data)
        # @Mstep [] correlate the fields to their columns in the model.
        self._correlate_to_columns()
        # @Mstep [] validate the schema fields.
        self.__validate()

        # data = {}
        # # @Mstep [LOOP] iterate this schema's fields
        # for f in self.fields:
        #     # @Mstep [IF] if the field is NOT dump_only
        #     if f.dump_only is False:
        #         # @Mstep [] add the field to the data dictionary.
        #         data[f.name] = f.value

        # @Mstep [] execute the insert on the model.
        self.model.insert(self).execute()
        # @Mstep [RETURN] return this schema instance.
        return self

    def update(self)->_t.schema_type:
        self._correlate_to_columns()
        self.__validate()
        data = {}
        for f in self.fields:
            if f.dump_only is False:
                data[f.name] = f.value
        self.model.update(**data).is_(self.model.primary_column.name,self.model.primary_column.value).execute()
        return self


    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} : {self.schema_name.s}>"


    def __setattr__(self, name, value):
        super().__setattr__(name, value)

        if not name in self._order:
            self._order.append(name)

        return value

    def ordered_attrs(self, with_order=False):
        '''Get a list of attributes for this schema in the order in which they were declared.'''
        return [(k,getattr(self, k)) for k in self._order if k != '_order' or with_order]



    @property
    def _unique_prop_keys(self):
        '''
            Get this Model's _unique_prop_keys

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 04-11-2023 15:45:07
            `@memberOf`: Model
            `@property`: _unique_prop_keys
        '''
        value = self.__unique_prop_keys
        if value is None:
            df_props = dir(Schema())
            # # @Mstep [] gather the props of this instance.
            props = dir(self)
            # # @Mstep [] find the props that exist on this instance and not on the base.
            value = c.arr.find_list_diff(props,df_props)
            self.__unique_prop_keys = value

        return value


    def _remove_missing_fields(self,data:dict):
        out = {}
        data = c.obj.keys_to_snake_case(data)
        for f in self.fields:
            key = f.name_in_dict(data)
            if key is not False:
                out[f.name] = data[key]
            else:
                f.not_provided = True
        return out

    def _correlate_to_dict(self,data:dict,ignore_unprovided=False):
        out_data = {}
        data = c.obj.keys_to_snake_case(data)
        # print(f"_correlate_to_dict: data:{data}")
        for f in self.fields:
            # snake_field = c.string.to_snake_case(f.name)
            # print(f"_correlate_to_dict: snake_field:{snake_field}")
            # if self.schema_crud_type == "update":
            #     # print(f"setting schema for update settings.")
            #     for f in self.fields:


            key = f.name_in_dict(data)
            if key is not False:
                out_data[f.name] = data[key]
                f.value = data[key]
                continue
            if key is False:
                out_data[f.name] = None
                # out_data[f.name] = _t.undefined
                f.value= None
                if self.schema_crud_type in ['update']:
                    f._ignore_null = True
                    f.default = None
                continue
            # if f.name in data:

            #     # out_data[f.name] = f.data_type_instance.deserialized_value(data[f.name])
            #     out_data[f.name] = data[f.name]
            #     f.value = data[f.name]
            #     continue
            if ignore_unprovided is False:
                if f.default != _t.no_default:
                    f.value = f.default
                    out_data[f.name] = f.default
            # @Mstep [IF] if the field is located in the path
            if f._open_api_location == "path":
                # @Mstep [IF] if the field has a default value.
                if f.default != _t.no_default:
                    # @Mstep [] set the fields value to its default
                    f.value = f.default
                    out_data[f.name] = f.default


        return out_data

    def _correlate_to_columns(self):
        '''
            Iterate all fields to find their corresponding columns in the current model.


            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 04-21-2023 07:40:42
            `memberOf`: Schema
            `version`: 1.0
            `method_name`: _correlate_to_columns
            * @xxx [04-21-2023 07:41:38]: documentation for _correlate_to_columns
        '''
        if self.model is None:
            raise ValueError(f"No Model Provided to {self.schema_name}")
        for f in self.fields:
            if isinstance(f,_nested_field):
                pri_col = self.model.primary_column
                child_model = _settings.globe.Volent.get_model(f.child_table)
                if child_model is not None:
                    child_model.select().is_(column_name=pri_col.name,value=pri_col.value)

            # print(f"{self.__class__.__name__} - self.model:{self.model}")
            col = self.model.get_column(f.name)
            if col is not None:
                f.column = col




# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=line-too-long
# pylint: disable=unused-import

# from dataclasses import dataclass
# from typing import Iterable, OrderedDict, Union
# from collections import OrderedDict
# from abc import ABC, abstractmethod
import re

import colemen_utils as c

import volent.settings as _settings
import volent.settings.types as _t
from volent.exceptions import ValidationError
from volent.validate.Validator import Validator

class LessThan(Validator):

    field_name = None
    default_message = "Must be Less than {comparable}"


    def __init__(
        self,
        comparable,
        inclusive=False,
        ):
        '''
        Validate that the comparable is less than the value.

        Arguments
        ----------------

        `comparable` {int,float}
            The value must be less than (or equal to) this value

        `inclusive` {bool} = False
            If True the value can be less than or eqaul to the comparable.
        '''
        self.comparable = comparable
        self.inclusive = inclusive


    def _format_error(self,value) -> str:
        if self.field_name is not None:
            if self.inclusive is True:
                return  f"{self.field_name} must be less than or equal to {self.comparable}"
            return  f"{self.field_name} must be less than {self.comparable}"
        return self.default_message.format(input=value, comparable=self.comparable)

    def _repr_args(self) -> str:
        return f"comparable={self.comparable!r}"

    def __call__(self, value: str,field_name:str=None) -> str:
        if value is None:
            return value
        self.field_name = field_name
        if self.inclusive is True:
            if value > self.comparable:
                raise ValidationError(self._format_error(value))
        else:
            if value >= self.comparable:
                raise ValidationError(self._format_error(value))

        return value


# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=line-too-long
# pylint: disable=unused-import

from dataclasses import dataclass
from typing import Any
from collections import OrderedDict
from abc import ABC, abstractmethod

import colemen_utils as c

import volent.settings as _settings
import volent.settings.types as _t
from volent.Column import Column as _column
from volent.Relationship import Relationship as _relationship
# from volent.mixins.EntityNameMixin import EntityNameMixin
# from volent.meta_validators import validate_table_comment,validate_table_name

class Validator(ABC):
    """Abstract base class for validators.

    .. note::
        This class does not provide any validation behavior. It is only used to
        add a useful `__repr__` implementation for validators.
    """

    error = None  # type: str | None

    def __repr__(self) -> str:
        args = self._repr_args()
        args = f"{args}, " if args else ""

        return "<{self.__class__.__name__}({args}error={self.error!r})>".format(
            self=self, args=args
        )

    def _repr_args(self) -> str:
        """A string representation of the args passed to this validator. Used by
        `__repr__`.
        """
        return ""

    @abstractmethod
    def __call__(self, value: Any) -> Any:
        ...


# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=line-too-long
# pylint: disable=unused-import

'''
    The master control settings for volent

    ----------

    Meta
    ----------
    `author`: Colemen Atwood
    `created`: 03-24-2023 07:16:43
    `name`: control
    * @xxx [03-24-2023 07:16:59]: documentation for control
'''




from typing import Union

open_api_param_locations = ["body","path"]


sql_quote_char:str = "`"
'''The character to use for quoting strings in the sql output'''
varchar_default_length:int = 50
'''The default character length for varchar columns'''

class query:
    '''Settings which are applied to all queries'''

    class pagination:
        items_per_page:int = 100
        default_page_number:int = 1

# mysql_maximum_column_name_length:int = 64
class mysql:
    '''MySQL specific settings'''
    create_table_indent:int = 4
    '''How many spaces to indent the contents of a create table statement'''
    safe_scripts:bool = True
    '''If True the statements include "IF [NOT] EXISTS" to avoid errors'''
    
    # _max_col_name_len:int = 64
    # '''The maximum length of a column name (don't fuckin change this, this is the MySQL limitation)'''
    _max_table_name_len:int = 64
    '''The maximum length of a table name (don't fuckin change this, this is the MySQL limitation)'''
    _max_database_name_len:int = 64
    '''The maximum length of a database name (don't fuckin change this, this is the MySQL limitation)'''
    _max_index_name_len:int = 64
    '''The maximum length of a index name (don't fuckin change this, this is the MySQL limitation)'''
    _max_constraint_name_len:int = 64
    '''The maximum length of a constraint name (don't fuckin change this, this is the MySQL limitation)'''

    _max_table_column_count:int = 4096
    '''The maximum number of columns a table can have (don't fuckin change this, this is the MySQL limitation)'''
    _max_table_comment_len:int = 2048
    '''The maximum length of a table comment (don't fuckin change this, this is the MySQL limitation)'''
    # _max_column_comment_len:int = 1024
    # '''The maximum length of a column comment (don't fuckin change this, this is the MySQL limitation)'''

    _reserved_terms = ["accessible","add","all","alter","analyze","and","as","asc","asensitive","before","between","bigint","binary","blob","both","by","call","cascade","case","change","char","character","check","collate","column","condition","constraint","continue","convert","create","cross","cube","cume_dist","current_date","current_time","current_timestamp","current_user","cursor","database","databases","day_hour","day_microsecond","day_minute","day_second","dec","decimal","declare","default","delayed","delete","dense_rank","desc","describe","deterministic","distinct","distinctrow","div","double","drop","dual","each","else","elseif","empty","enclosed","escaped","except","exists","exit","explain","false","fetch","first_value","float","float4","float8","for","force","foreign","from","fulltext","function","generated","get","grant","group","grouping","groups","having","high_priority","hour_microsecond","hour_minute","hour_second","if","ignore","in","index","infile","inner","inout","insensitive","insert","int","int1","int2","int3","int4","int8","integer","intersect","interval","into","io_after_gtids","io_before_gtids","is","iterate","join","json_table","key","keys","kill","lag","last_value","lateral","lead","leading","leave","left","like","limit","linear","lines","load","localtime","localtimestamp","lock","long","longblob","longtext","loop","low_priority","master_bind","master_ssl_verify_server_cert","match","maxvalue","mediumblob","mediumint","mediumtext","middleint","minute_microsecond","minute_second","mod","modifies","natural","not","no_write_to_binlog","nth_value","ntile","null","numeric","of","on","optimize","optimizer_costs","option","optionally","or","order","out","outer","outfile","over","partition","percent_rank","precision","primary","procedure","purge","range","rank","read","reads","read_write","real","recursive","references","regexp","release","rename","repeat","replace","require","resignal","restrict","return","revoke","right","rlike","row","rows","row_number","schema","schemas","second_microsecond","select","sensitive","separator","set","show","signal","smallint","spatial","specific","sql","sqlexception","sqlstate","sqlwarning","sql_big_result","sql_calc_found_rows","sql_small_result","ssl","starting","stored","straight_join","system","table","terminated","then","tinyblob","tinyint","tinytext","to","trailing","trigger","true","undo","union","unique","unlock","unsigned","update","usage","use","using","utc_date","utc_time","utc_timestamp","values","varbinary","varchar","varcharacter","varying","virtual","when","where","while","window","with","write","xor","year_month","zerofill"]
    '''A list of terms that are reserved by MySQL (don't fuckin change this, this is the MySQL limitation)'''
    _invalid_name_first_chars = [*"@#$0123456789"]
    '''A list of characters that a column name cannot start with. (don't fuckin change this, this is the MySQL limitation)'''

    ignore_reserved_terms:bool = False
    '''If True, the name validation will not raise an error if there is a collision.'''




    class database:
        max_name_length:int = 64
        '''The maximum length of a database name (don't fuckin change this, this is the MySQL limitation)'''


    class table:

        max_comment_length:int = 2048
        '''The maximum length of a table comment (don't fuckin change this, this is the MySQL limitation)'''
        max_name_length:int = 64
        '''The maximum length of a table name (don't fuckin change this, this is the MySQL limitation)'''
        _max_column_count:int = 4096
        '''The maximum number of columns a table can have (don't fuckin change this, this is the MySQL limitation)'''


    class column:

        max_comment_length:int = 1024
        '''The maximum length of a column comment (don't fuckin change this, this is the MySQL limitation)'''
        max_name_length:int = 64
        '''The maximum length of a column name (don't fuckin change this, this is the MySQL limitation)'''









# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=line-too-long
# pylint: disable=unused-import

from dataclasses import dataclass
from typing import Iterable
import colemen_utils as c
import volent.settings as _settings
import volent.settings.types as _t
from volent.data_types.TypeBase import TypeBase as _type_base
from volent.exceptions import ValidationError


@dataclass
class Bool(_type_base):


    def __init__(self) -> None:
        super().__init__()
        self.sql_type_name = "BOOLEAN"
        self.python_data_type = (bool)
        self.open_api_data_type = "boolean"


    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} >"


    def serializer(self,value,field_name:str=None):

        if isinstance(value,(bool)) is False:
            return self.deserialized_value(value,field_name)

        else:
            return self.serialized_value(value,field_name)



    def serialized_value(self,value,field_name:str=None):

        if isinstance(value,(bool)):
            return value

        value = c.types.bool_to_int(value)

        if isinstance(value,(int)) is False:
            raise ValidationError(f"Failed to serialize {value} to an integer.",field_name)

        return value

    def deserialized_value(self,value,field_name:str=None):

        if isinstance(value,(bool)):
            return value
        if value == _t.undefined or value == _t.no_default:
            return value
        if value is None:
            return value

        value = c.types.to_bool(value)

        if isinstance(value,(bool)) is False:
            raise ValidationError(f"Failed to serialize {value} to a boolean.",field_name)

        return value

    


# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=line-too-long
# pylint: disable=unused-import

from dataclasses import dataclass
from typing import Iterable, OrderedDict, Sequence, Sized, Any, TypeVar
from collections import OrderedDict
from abc import ABC, abstractmethod


import colemen_utils as c

import volent.settings as _settings
import volent.settings.types as _t
from volent.exceptions import ValidationError
from volent.validate.Validator import Validator


_T = TypeVar("_T")

class OneOf(Validator):

    field_name = None
    default_message = "One or more of the choices you made was not in: {choices}."


    def __init__(
        self,
        choices=Iterable,
        ):

        self.choices = choices
        self.choices_str = ', '.join(str(choice) for choice in self.choices)

    def _format_error(self, value: Any, message: str) -> str:
        if self.field_name is not None:
            message = f"{self.field_name} [{value}] was not in {self.choices}"
        return message.format(
            input=value, choices=self.choices_str
        )


    def _repr_args(self) -> str:
        return f"min={self.min!r}, max={self.max!r}, equal={self.equal!r}"


    def __call__(self,value:Any,field_name:str=None)->Any:
        self.field_name = field_name
        test_value = c.arr.force_list(value)
        match_found = False
        for val in test_value:
            if val in self.choices:
                match_found = True
        if match_found is False:
            raise ValidationError(self._format_error(value,self.default_message))
        return value


class ContainsOnly(OneOf):

    field_name = None
    default_message = "One or more of the choices you made was not in: {choices}."


    def __init__(
        self,
        choices=Iterable,
        ):

        self.choices = choices
        self.choices_str = ', '.join(str(choice) for choice in self.choices)

    def _format_error(self, value: str, message: str) -> str:
        if self.field_name is not None:
            message = f"{self.field_name} was not in {self.choices}"
        return message.format(
            input=value, choices=self.choices_str
        )


    def _repr_args(self) -> str:
        return f"min={self.min!r}, max={self.max!r}, equal={self.equal!r}"


    def __call__(self,value:Sequence[_T],field_name:str=None)->Sequence[_T]:
        self.field_name = field_name
        test_value = c.arr.force_list(value)

        for val in test_value:
            if val not in self.choices:
                raise ValidationError(self._format_error(value,self.default_message))

        return value
